#!/usr/bin/env python2
# coding: utf-8
# Distributed under the terms of the MIT License.


""" This script uses qmpy to loop over all entries on a
self-hosted MySQL OQMD database and creates a matador
MongoDB collection containing entries in the matador format.

Instructions for self-hosting the OQMD can be found at:

http://oqmd.org/static/docs/getting_started.html#setting-up-the-database

This script has only been tested with MariaDB 10 (with Centos 7 packages
mariadb-devel, mariadb-server), which are required to install qmpy.

Note:
    As of writing this script, qmpy is Python 2 only, so is not compatible
    with the main matador core.

This script can balloon in memory usage in restart mode. You may wish to
invoke it as `while ./spatulate_oqmd.py --append --no_restart; do :; done;`
to clear the memory.

"""

from random import randint
from os.path import dirname, realpath
import argparse
from traceback import print_exc
import numpy as np
import pymongo as pm
import tqdm
import qmpy


def qmpy_parse_kpoints(string):
    """ Parse VASP kpoint block into grid (ignores offset). """
    return [int(i) for i in string.split('\n')[-2].split(' ')]


def qmpy_entry_to_doc(entry):
    """ Parse a qmpy entry object into a matador document,
    with associated VASP parameters converted into the CASTEP
    keywords.

    Parameters:
        entry (:obj:`qmpy.Entry`): the entry in the OQMD.

    Returns:
        dict: matador document.

    """

    doc = dict()

    if entry is None:
        raise RuntimeError('Entry is None.')
    if entry.duplicate_of is not None and entry.duplicate_of.id != entry.id:
        raise RuntimeError('Structure is a duplicate.')

    doc['task'] = 'geometryoptimization'

    try:
        doc['num_atoms'] = int(entry.natoms)
    except TypeError:
        doc['num_atoms'] = int(entry.structure.natoms)

    doc['elems'] = sorted(list({elem.symbol for elem in entry.elements}))
    doc['stoichiometry'] = sorted([[elem, entry.comp[elem]] for elem in entry.comp], key=lambda x: x[0])
    doc['num_fu'] = doc['num_atoms'] / int(sum(doc['stoichiometry'][i][1] for i in range(len(doc['stoichiometry']))))
    try:
        doc['formation_energy_per_atom'] = entry.energy
        doc['formation_energy'] = entry.energy * doc['num_atoms']
    except TypeError:
        raise RuntimeError('entry.id={} has no energy'.format(entry.id))
    doc['enthalpy_per_atom'] = entry.energy
    doc['enthalpy'] = entry.energy * doc['num_atoms']
    doc['source'] = ['OQMD {}'.format(int(entry.id))]
    doc['_oqmd_entry_id'] = int(entry.id)
    if 'icsd' in entry.keywords:
        doc['icsd'] = int(entry.label.split('-')[-1])
        try:
            doc['reference'] = str(entry.reference.citation)
        except Exception:
            pass

    doc['space_group'] = entry.spacegroup.hm
    doc['lattice_cart'] = entry.structure.cell.tolist()
    doc['positions_abs'] = entry.structure.cartesian_coords.tolist()
    doc['positions_frac'] = entry.structure.coords.tolist()
    doc['atom_types'] = []
    for atom in entry.structure.atoms:
        doc['atom_types'].append(atom.species)
    doc['cell_volume'] = entry.structure.volume
    doc['task'] = 'geometryoptimization'
    doc['species_pot'] = {elem: 'OQMD' for elem in doc['elems']}
    doc['stable'] = entry.stable

    # some entries have multiple E_f, choose the last one,
    # which should match entry.energy: OR SO I THOUGHT
    # instead, we need to loop over the formation energies and find
    # the matching one. This is potentially the lowest one every time.
    efs = entry.formationenergy_set.all()
    for ef in efs:
        if abs(ef.delta_e - entry.energy) < 1e-6:
            eform = ef
            break
    else:
        print('Unable to match formation energy for entry.id={}'.format(entry.id))
        raise RuntimeError
    doc['hull_distance'] = max(0, eform.stability)
    ef_calc = eform.calculation
    oqmd_calc_settings = ef_calc.settings
    doc['cut_off_energy'] = oqmd_calc_settings['encut']
    doc['spin_polarized'] = bool(oqmd_calc_settings['ispin'] - 1)
    doc['kpoints_mp_grid'] = qmpy_parse_kpoints(ef_calc.KPOINTS)
    doc['kpoints_mp_spacing'] = calc_mp_spacing(doc['lattice_cart'], doc['kpoints_mp_grid'])
    doc['xc_functional'] = 'PBE'

    doc['forces'] = ef_calc.output.forces
    doc['pressure'] = 0.0
    doc['stress'] = -0.1 * (ef_calc.output.sxx + ef_calc.output.syy + ef_calc.output.szz)/3.0
    try:
        doc['max_force_on_atom'] = np.max(np.linalg.norm(doc['forces'], axis=-1))
    except Exception:
        print(entry.id, doc['forces'])
    doc['forces'] = doc['forces'].tolist()

    return doc


def calc_mp_spacing(real_lat, mp_grid, prec=3):
    """ Convert real lattice in Cartesian basis and the
    kpoint_mp_grid into a grid spacing. Copied from matador
    utils version.

    Parameters:
        real_lat (list): Cartesian lattice vectors.
        mp_grid (:obj:`list` of :obj:`int`): 3 integers defining the MP grid.

    Keyword arguments:
        prec (int): desired decimal precision of output.

    Returns:
        float: mp_spacing rounded to `prec`.

    """
    recip_lat = real2recip(real_lat)
    recip_len = np.zeros((3))
    recip_len = np.sqrt(np.sum(np.power(recip_lat, 2), axis=1))
    spacing = recip_len / (2*np.pi*np.asarray(mp_grid))
    max_spacing = np.max(spacing)
    exponent = round(np.log10(max_spacing) - prec)
    return round(max_spacing + 0.5*10**exponent, prec)


def real2recip(real_lat):
    """ Convert the real lattice in Cartesian basis to
    the reciprocal space lattice. Copy of matador.utils
    version.

    Parameters:
        real_lat (list): Cartesian lattice vectors.

    Returns:
        list: Cartesian lattice vectors of reciprocal lattice.

    """
    real_lat = np.asarray(real_lat)
    recip_lat = np.zeros((3, 3))
    recip_lat[0] = (2*np.pi)*np.cross(real_lat[1], real_lat[2]) / \
        (np.dot(real_lat[0], np.cross(real_lat[1], real_lat[2])))
    recip_lat[1] = (2*np.pi)*np.cross(real_lat[2], real_lat[0]) / \
        (np.dot(real_lat[1], np.cross(real_lat[2], real_lat[0])))
    recip_lat[2] = (2*np.pi)*np.cross(real_lat[0], real_lat[1]) / \
        (np.dot(real_lat[2], np.cross(real_lat[0], real_lat[1])))
    return recip_lat.tolist()


class DBConverter:
    def __init__(
        self,
        host=None, client=None, dryrun=False, debug=False,
        verbosity=0, db_name='oqmd_1.2', append=False,
        start_id=0, chunk_size=1000, restart=True
    ):
        """ Connect to the relevant databases and
        set off the scraper.
        """
        self.import_count = 0
        self.success_count = 0
        self.limit = 50000
        self.num_scraped = 0
        self.dryrun = dryrun
        self.debug = debug
        self.chunk_size = chunk_size
        self.db_name = db_name
        self.client = client
        self.verbosity = verbosity
        self.append = append
        self.restart = restart
        self.start_id = start_id
        # set up I/O for text_id
        if not self.dryrun:
            print(__file__)
            wordfile = open(dirname(realpath(__file__)) + '/../../matador/scrapers/words', 'r')
            nounfile = open(dirname(realpath(__file__)) + '/../../matador/scrapers/nouns', 'r')
            self.wlines = wordfile.readlines()
            self.num_words = len(self.wlines)
            self.nlines = nounfile.readlines()
            self.num_nouns = len(self.nlines)
            wordfile.close()
            nounfile.close()

            if self.client is None:
                print('connecting to {}'.format(host))
                self.client = pm.MongoClient(host)

            self.db = self.client.crystals
            current_collections = self.db.list_collection_names()
            if self.db_name in current_collections and not self.append:
                raise SystemExit('Desired db_name already exists!')

            self.repo = self.db[self.db_name]
            self.create_indices()
            self.build_mongo()

    def create_indices(self):
        # create unique index for oqmd ID to allow for repeated imports
        self.repo.create_index([('enthalpy_per_atom', pm.ASCENDING)])
        self.repo.create_index([('stoichiometry', pm.ASCENDING)])
        self.repo.create_index([('elems', pm.ASCENDING)])
        self.repo.create_index([('source', pm.ASCENDING)])
        self.create_extra_indices()

    def struct2db(self, struct):
        """ Insert completed Python dictionary into chosen
        database, with generated text_id.
        """
        plain_text_id = [self.wlines[randint(0, self.num_words-1)].strip(),
                         self.nlines[randint(0, self.num_nouns-1)].strip()]
        struct['text_id'] = plain_text_id
        if '_id' in struct:
            raise RuntimeError('{}'.format(struct))
        try:
            _ = self.repo.insert_one(struct)
        except pm.errors.DuplicateKeyError:
            return 0

        return 1


class MPConverter(DBConverter):

    def __init__(self, cursor, *args, **kwargs):
        self.cursor = cursor
        super().__init__(*args, **kwargs)

    def create_extra_indices(self):
        # create unique index for MP ID to allow for repeated imports
        self.repo.create_index('_mp_id', unique=True)

    def build_mongo(self):
        for doc in self.cursor:
            doc = self.doc2entry(doc)
            self.import_count += self.struct2db(doc)

        print(self.import_count, '/', len(self.cursor))

    def doc2entry(self, doc):
        if 'task' not in doc:
            doc['task'] = 'geometryoptimization'
        if 'species_pot' not in doc:
            doc['species_pot'] = {elem: 'OQMD' for elem in doc['elems']}
        for key in doc:
            if isinstance(doc[key], np.ndarray):
                doc[key] = doc[key].tolist()
            elif isinstance(doc[key], set):
                doc[key] = sorted(list(doc[key]))

        return doc


class OQMDConverter(DBConverter):
    """ The OQMDConverter class implements methods to scrape
    the OQMD SQL database for all entries using the qmpy interface.
    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def create_extra_indices(self):
        # create unique index for oqmd ID to allow for repeated imports
        self.repo.create_index('_oqmd_entry_id', unique=True)

    def build_mongo(self):
        """ Perform QMPY query for all entries, and scrape them into a MongoDB. """
        # start by scraping all converged structures with chosen label
        chunk_iter = 0
        all_structures = qmpy.Entry.objects.all().count()
        print('Expecting {} structures total'.format(all_structures))

        try:
            self.start_id = self.repo.find().sort('_oqmd_entry_id', pm.DESCENDING).limit(1)[0]['_oqmd_entry_id']
            print('Restarting from entry.id={}'.format(self.start_id))
        except IndexError:
            self.start_id = 0

        while self.num_scraped < all_structures and self.import_count < self.limit:
            chunk_min = self.start_id + chunk_iter * self.chunk_size
            chunk_max = chunk_min + self.chunk_size
            chunk_iter += 1

            # check that there are even any structures left to find
            remaining = qmpy.Entry.objects.filter(id__gte=chunk_min).count()
            print('{} structures remaining with ID > {}'.format(remaining, chunk_min))
            if remaining == 0:
                break

            cursor = (qmpy.Entry.objects
                      .filter(id__gte=chunk_min)
                      .filter(id__lt=chunk_max))

            num_structures = cursor.count()
            if num_structures < 1:
                continue

            print('Chunk {} -> {} contains {} entries'.format(chunk_min, chunk_max, num_structures))
            for entry in tqdm.tqdm(cursor, total=num_structures):
                try:
                    doc = {}
                    doc = qmpy_entry_to_doc(entry)
                except RuntimeError:
                    print_exc()
                    continue
                except Exception:
                    print_exc()
                    continue

                self.success_count += 1
                if not self.dryrun:
                    self.import_count += self.struct2db(doc)

            # seems like this keeps memory down in Python 2.7, otherwise it can balloon
            del cursor

            self.num_scraped = self.repo.count_documents({})
            print('Reached {} de-duplicated entiries out of {} including duplicates'.format(self.num_scraped, all_structures))

            if not self.restart:
                break

        if self.dryrun:
            print('Successfully scraped', self.success_count, '/',
                  'structures.')
        if not self.dryrun:
            if self.import_count == 0:
                raise RuntimeError('Nothing imported.')
            print('Successfully imported', self.import_count, '/',
                  'structures.')


if __name__ == '__main__':
    # importer = QMPYConverter()
    parser = argparse.ArgumentParser(
        description='Import OQMD (http://oqmd.org) structures into MongoDB database.',
        epilog='Written by Matthew Evans (2016)')
    parser.add_argument('-d', '--dryrun', action='store_true',
                        help='run the importer without connecting to the database')
    parser.add_argument('-v', '--verbosity', action='count',
                        help='enable verbose output')
    parser.add_argument('--no_restart', action='store_true',
                        help='don\'t restart script if there are missing entries')
    parser.add_argument('--debug', action='store_true',
                        help='enable debug output to print every dict')
    parser.add_argument('--append', action='store_true',
                        help='add to existing collection')
    args = parser.parse_args()
    importer = OQMDConverter(dryrun=args.dryrun,
                             debug=args.debug,
                             db_name='oqmd_1.2',
                             restart=not args.no_restart,
                             append=args.append,
                             verbosity=args.verbosity)



# coding: utf-8
# Distributed under the terms of the MIT License.

""" This file implements convex hull functionality from database
queries.

"""


from copy import deepcopy
from typing import List
from collections import defaultdict
import re
import warnings

from bson.son import SON
import pymongo as pm
import numpy as np

from matador.utils.print_utils import print_notify, print_warning
from matador.utils.chem_utils import parse_element_string, get_padded_composition, get_num_intercalated
from matador.utils.chem_utils import get_generic_grav_capacity, get_formula_from_stoich, get_stoich_from_formula
from matador.utils.chem_utils import get_formation_energy
from matador.utils.cursor_utils import set_cursor_from_array, get_array_from_cursor
from matador.utils.cursor_utils import display_results, recursive_get
from matador.utils.cursor_utils import filter_cursor_by_chempots
from matador.battery import Electrode, VoltageProfile
from matador.hull.phase_diagram import PhaseDiagram

# general small number used when comparing energies to zero
EPS = 1e-8
# small number used for checking boundaries
BOUNDARY_EPS = 1e-12


class QueryConvexHull:
    """ Construct a binary or ternary phase diagram from a
    matador.query.DBQuery object, or a list of structures.

    Attributes:
        cursor (list): list of all structures used to create phase diagram.
        hull_cursor (list): list of all documents within hull_cutoff.
        chempot_cursor (list): list of chemical potential documents.
        structures (numpy.ndarray): all structures used to create hull.
        hull_dist (np.ndarray): array of distances from hull for each structure.
        species (list): list of chemical potential symbols.
        num_elements (int): number of elements present in the chemical potentials.
        elements (list): the elements present in the convex hull.
        voltage_data (list): if voltage_curve() has been called, then
            this is a lsit of VoltageProfile objects containing Q, x, V and reaction pathways.
        volume_data (dict): if volume_curve() has been called, then
            this is a dictionary containing x and volumes per B (in AxBy).

    """

    def __init__(self, query=None, cursor=None, elements=None, species=None, voltage=False, volume=False, subcmd=None,
                 plot_kwargs=None, lazy=False, energy_key='enthalpy_per_atom', client=None, collections=None, db=None,
                 **kwargs):
        """ Initialise the class from either a DBQuery or a cursor (list
        of matador dicts) and construct the appropriate phase diagram.

        Keyword arguments:
            query (matador.query.DBQuery): object containing structures,
            cursor (list(dict)): alternatively specify list of matador documents.
            species (list(str)): list of elements/chempots to use, used to provide a useful order,
            voltage (bool): whether or nto to compute voltages relative for insertion of first entry in species,
            volume (bool): whether or not to compute volume expansion relative to first entry in species,
            energy_key (str): key under which the desired energy *per atom* is stored.
            lazy (bool): if True, do not create hull until `self.create_hull()` is called
            chempots (list(float)): list of chemical potential values to use.
            elements (list(str)): deprecated form `species`.
            kwargs (dict): mostly CLI arguments, see matador hull --help for full options.
            plot_kwargs (dict): arguments to pass to plot_hull function
            client (pymongo.MongoClient): optional client to pass to DBQuery.
            collections (dict of pymongo.collections.Collection): optional dict of collections to pass to DBQuery.
            db (str): db name to connect to in DBQuery.

        """
        self.args = dict()
        if query is not None:
            self.args.update(query.args)
        self.args.update(kwargs)

        if subcmd is not None:
            warnings.warn("subcmd will soon be deprecated, please pass the equivalent flag as a kwarg (e.g. voltage=True)")
            if subcmd == 'voltage':
                voltage = True
        self.args['subcmd'] = subcmd

        if plot_kwargs is None:
            plot_kwargs = {}
        self.args['plot_kwargs'] = plot_kwargs

        self.from_cursor = False
        self.plot_params = False

        self.compute_voltages = voltage
        self.compute_volumes = volume

        if query is None and cursor is None:
            # if no query or cursor passed, push all kwargs to new query
            from matador.query import DBQuery
            kwargs.pop('intersection', None)
            query = DBQuery(
                subcmd='hull',
                intersection=True,
                client=client,
                collections=collections,
                **kwargs
            )

        self._query = query

        if self._query is not None:
            # this isn't strictly necessary but it maintains the sanctity of the query results
            self.cursor = list(deepcopy(query.cursor))
            self.args['use_source'] = False
            if self._query.args['subcmd'] not in ['hull', 'voltage', 'hulldiff']:
                print_warning('Query was not prepared with subcmd=hull, so cannot guarantee consistent formation energies.')
        else:
            self.cursor = list(cursor)
            self.from_cursor = True
            self.args['use_source'] = True

        # set up attributes for later
        self.structures = None
        self.convex_hull = None
        self.chempot_cursor = None
        self.hull_cursor = None
        self.phase_diagram = None
        self.hull_dist = None
        self.species = None
        self.voltage_data: List[VoltageProfile] = []
        self.volume_data = defaultdict(list)
        self.elements = []
        self.num_elements = 0
        self.species = self._get_species(species, elements)
        self._dimension = len(self.species)

        # tracker for whether per_b fields have been setup
        self._per_b_done = False

        self.energy_key = energy_key
        if not self.energy_key.endswith('_per_atom'):
            warnings.warn('Appending per_atom to energy_key {}'.format(self.energy_key))
            self.energy_key += '_per_atom'

        self._extensive_energy_key = self.energy_key.split('_per_atom')[0]

        if self.args.get('hull_cutoff') is not None:
            self.hull_cutoff = float(self.args['hull_cutoff'])
        else:
            self.hull_cutoff = 0.0

        if self.cursor is None:
            raise RuntimeError('Failed to find structures to create hull!')

        self._non_elemental = False
        assert isinstance(self.species, list)
        for _species in self.species:
            if len(parse_element_string(_species, stoich=True)) > 1:
                self._non_elemental = True

        if not lazy:
            self.create_hull()

    def create_hull(self):
        """ Begin the hull creation routines and perform the
        post-processing specified by initial arguments.

        """
        if self.args.get('uniq'):
            from matador.utils.cursor_utils import filter_unique_structures
            if self.args.get('uniq') is True:
                sim_tol = 0.1
            else:
                sim_tol = self.args.get('uniq')
            print_notify('Filtering for unique structures...')
            self.cursor = filter_unique_structures(
                self.cursor,
                args=self.args,
                quiet=True,
                sim_tol=sim_tol,
                hull=True,
                energy_key=self.energy_key
            )

        self.construct_phase_diagram()

        if not self.hull_cursor:
            print_warning('No structures on hull with chosen chemical potentials.')
        else:
            print_notify('{} structures found within {} eV of the hull, including chemical potentials.'
                         .format(len(self.hull_cursor), self.hull_cutoff))

        display_results(self.hull_cursor, hull=True, energy_key=self.energy_key, **self.args)

        if self.compute_voltages:
            print("Constructing electrode system with active ion: {}".format(self.species[0]))
            self.voltage_curve([doc for doc in self.hull_cursor if doc['hull_distance'] <= 1e-9])

        if self.compute_volumes:
            self.volume_curve()

        if not self.args.get('no_plot'):
            if self.compute_voltages and self.voltage_data:
                print('plotting voltage')
                self.plot_voltage_curve(show=False)
            if self.compute_volumes and self.volume_data:
                self.plot_volume_curve(show=False)

            self.plot_hull(**self.args['plot_kwargs'], debug=self.args.get('debug'), show=True)

    def __repr__(self):
        return display_results(self.hull_cursor, args=self.args, hull=True, energy_key=self.energy_key, return_str=True)

    @property
    def savefig(self):
        """ True if any figure type argument was passed. """
        return any([self.args.get('pdf'), self.args.get('png'), self.args.get('svg')])

    @property
    def savefig_ext(self):
        """ Returns the figure extension to save with. """
        for ext in ['pdf', 'png', 'svg']:
            if self.args.get(ext):
                return ext

    def plot_hull(self, **kwargs):
        """ Hull plot helper function. """
        from matador import plotting
        if self._dimension == 3:
            ax = plotting.plot_ternary_hull(self, **kwargs)
        elif self._dimension == 2:
            ax = plotting.plot_2d_hull(self, **kwargs)
        else:
            print_notify('Unable to plot phase diagram of dimension {}.'.format(self._dimension))
        return ax

    def plot_voltage_curve(self, **kwargs):
        """ Voltage plot helper function. """
        from matador import plotting
        savefig = None
        if self.savefig:
            savefig = '{}_voltage.{}'.format(''.join(self.elements), self.savefig_ext)

        plotting.plot_voltage_curve(
            self.voltage_data,
            expt=self.args.get('expt'),
            savefig=savefig,
            labels=self.args.get('labels'),
            expt_label=self.args.get('expt_label'),
            **kwargs
        )

    def plot_volume_curve(self, **kwargs):
        """ Volume plot helper function. """
        from matador import plotting
        plotting.plot_volume_curve(
            self,
            **self.args['plot_kwargs'],
            show=False
        )

    def fake_chempots(self, custom_elem=None):
        """ Spoof documents for command-line chemical potentials.

        Keyword arguments:
            custom_elem (list(str)): list of element symbols to generate chempots for.

        """
        self.chempot_cursor = []
        print('Generating fake chempots...')

        if custom_elem is None:
            custom_elem = self.species

        if len(custom_elem) != len(self.species):
            raise RuntimeError('Wrong number of compounds/chemical potentials specified: {} vs {}'
                               .format(custom_elem, self.args.get('chempots')))
        for i, _ in enumerate(self.args.get('chempots')):
            self.chempot_cursor.append(dict())
            self.chempot_cursor[i]['stoichiometry'] = get_stoich_from_formula(custom_elem[i])
            self.chempot_cursor[i][self.energy_key] = -1*abs(self.args.get('chempots')[i])
            self.chempot_cursor[i][self._extensive_energy_key] = self.chempot_cursor[i][self.energy_key] * \
                sum(elem[1] for elem in self.chempot_cursor[i]['stoichiometry'])
            self.chempot_cursor[i]['num_fu'] = 1
            self.chempot_cursor[i]['num_atoms'] = 1
            self.chempot_cursor[i]['text_id'] = ['command', 'line']
            self.chempot_cursor[i]['_id'] = None
            self.chempot_cursor[i]['source'] = ['command_line']
            self.chempot_cursor[i]['space_group'] = 'xxx'
            self.chempot_cursor[i][self._extensive_energy_key + '_per_b'] = self.chempot_cursor[i][self.energy_key]
            self.chempot_cursor[i]['num_a'] = 0
            self.chempot_cursor[i]['cell_volume'] = 1
            self.chempot_cursor[i]['concentration'] = [1 if i == ind else 0 for ind in range(self._dimension-1)]
        self.chempot_cursor[0]['num_a'] = float('inf')
        notify = 'Custom chempots:'
        for chempot in self.chempot_cursor:
            notify += '{:3} = {} eV/fu, '.format(get_formula_from_stoich(chempot['stoichiometry'], sort=False),
                                                 chempot[self._extensive_energy_key])

        if self.args.get('debug'):
            for match in self.chempot_cursor:
                print(match)
        print(len(notify) * '─')
        print(notify)
        print(len(notify) * '─')

    def construct_phase_diagram(self):
        """ Create a phase diagram with arbitrary chemical potentials.

        Expects self.cursor to be populated with structures and chemical potential
        labels to be set under self.species.

        """
        self.set_chempots()
        self.cursor = filter_cursor_by_chempots(self.species, self.cursor)

        formation_key = 'formation_{}'.format(self.energy_key)
        extensive_formation_key = 'formation_{}'.format(self._extensive_energy_key)
        for ind, doc in enumerate(self.cursor):
            self.cursor[ind][formation_key] = get_formation_energy(self.chempot_cursor, doc,
                                                                   energy_key=self.energy_key)
            self.cursor[ind][extensive_formation_key] = doc[formation_key] * doc['num_atoms']

        if self._non_elemental and self.args.get('subcmd') in ['voltage', 'volume']:
            raise NotImplementedError('Pseudo-binary/pseudo-ternary voltages not yet implemented.')

        self.phase_diagram = PhaseDiagram(self.cursor, formation_key, self._dimension)
        # aliases for data stored in phase diagram
        self.structures = self.phase_diagram.structures
        self.hull_dist = self.phase_diagram.hull_dist
        self.convex_hull = self.phase_diagram.convex_hull

        # ensure hull cursor is sorted by enthalpy_per_atom,
        # then by concentration, as it will be by default if from database
        hull_cursor = [self.cursor[idx] for idx in np.where(self.hull_dist <= self.hull_cutoff + EPS)[0]]
        # TODO: check why this fails when the opposite way around
        hull_cursor = sorted(hull_cursor, key=lambda doc: (doc['concentration'], recursive_get(doc, self.energy_key)))

        # by default hull cursor includes all structures within hull_cutoff
        # if summary requested and we're in hulldiff mode, filter hull_cursor for lowest per stoich
        if self.args.get('summary') and self.args['subcmd'] == 'hulldiff':
            tmp_hull_cursor = []
            compositions = set()
            for ind, member in enumerate(hull_cursor):
                formula = get_formula_from_stoich(member['stoichiometry'], sort=True)
                if formula not in compositions:
                    compositions.add(formula)
                    tmp_hull_cursor.append(member)

            hull_cursor = tmp_hull_cursor

        self.hull_cursor = hull_cursor

    def set_chempots(self, energy_key=None):
        """ Search for chemical potentials that match the structures in
        the query cursor and add them to the cursor. Also set the concentration
        of chemical potentials in :attr:`cursor`, if not already set.

        """
        if energy_key is None:
            energy_key = self.energy_key
        query = self._query
        query_dict = dict()
        species_stoich = [sorted(get_stoich_from_formula(spec, sort=False)) for spec in self.species]
        self.chempot_cursor = []

        if self.args.get('chempots') is not None:
            self.fake_chempots(custom_elem=self.species)

        elif self.from_cursor:
            chempot_cursor = sorted([doc for doc in self.cursor if doc['stoichiometry'] in species_stoich],
                                    key=lambda doc: recursive_get(doc, energy_key))

            for species in species_stoich:
                for doc in chempot_cursor:
                    if doc['stoichiometry'] == species:
                        self.chempot_cursor.append(doc)
                        break

            if len(self.chempot_cursor) != len(self.species):
                raise RuntimeError('Found {} of {} required chemical potentials'
                                   .format(len(self.chempot_cursor), len(self.species)))

            if self.args.get('debug'):
                print([mu['stoichiometry'] for mu in self.chempot_cursor])

        else:
            print(60 * '─')
            self.chempot_cursor = len(self.species) * [None]
            # scan for suitable chem pots in database
            for ind, elem in enumerate(self.species):

                print('Scanning for suitable', elem, 'chemical potential...')
                query_dict['$and'] = deepcopy(list(query.calc_dict['$and']))

                if not self.args.get('ignore_warnings'):
                    query_dict['$and'].append(query.query_quality())

                if len(species_stoich[ind]) == 1:
                    query_dict['$and'].append(query.query_composition(custom_elem=[elem]))
                else:
                    query_dict['$and'].append(query.query_stoichiometry(custom_stoich=[elem]))

                # if oqmd, only query composition, not parameters
                if query.args.get('tags') is not None:
                    query_dict['$and'].append(query.query_tags())

                mu_cursor = query.repo.find(SON(query_dict)).sort(energy_key, pm.ASCENDING)
                if mu_cursor.count() == 0:
                    print_notify('Failed... searching without spin polarization field...')
                    scanned = False
                    while not scanned:
                        for idx, dicts in enumerate(query_dict['$and']):
                            for key in dicts:
                                if key == 'spin_polarized':
                                    del query_dict['$and'][idx][key]
                                    break
                            if idx == len(query_dict['$and']) - 1:
                                scanned = True
                    mu_cursor = query.repo.find(SON(query_dict)).sort(energy_key, pm.ASCENDING)

                if mu_cursor.count() == 0:
                    raise RuntimeError('No chemical potentials found for {}...'.format(elem))

                self.chempot_cursor[ind] = mu_cursor[0]
                if self.chempot_cursor[ind] is not None:
                    print('Using', ''.join([self.chempot_cursor[ind]['text_id'][0], ' ',
                                            self.chempot_cursor[ind]['text_id'][1]]), 'as chem pot for', elem)
                    print(60 * '─')
                else:
                    raise RuntimeError('No possible chem pots available for {}.'.format(elem))

            for i, mu in enumerate(self.chempot_cursor):
                self.chempot_cursor[i][self._extensive_energy_key + '_per_b'] = mu[energy_key]
                self.chempot_cursor[i]['num_a'] = 0

            self.chempot_cursor[0]['num_a'] = float('inf')

        # don't check for IDs if we're loading from cursor
        if not self.from_cursor:
            ids = [doc['_id'] for doc in self.cursor]
            if self.chempot_cursor[0]['_id'] is None or self.chempot_cursor[0]['_id'] not in ids:
                self.cursor.insert(0, self.chempot_cursor[0])
            for match in self.chempot_cursor[1:]:
                if match['_id'] is None or match['_id'] not in ids:
                    self.cursor.append(match)

        # add faked chempots to overall cursor
        elif self.args.get('chempots') is not None:
            self.cursor.insert(0, self.chempot_cursor[0])
            self.cursor.extend(self.chempot_cursor[1:])

        # find all elements present in the chemical potentials
        elements = []
        for mu in self.chempot_cursor:
            for elem, _ in mu['stoichiometry']:
                if elem not in elements:
                    elements.append(elem)
        self.elements = elements
        self.num_elements = len(elements)

    @staticmethod
    def filter_cursor_by_chempots(species, cursor):
        """ For the desired chemical potentials, remove any incompatible structures
        from cursor.

        Parameters:
            species (list): list of chemical potential formulae.
            cursor (list): list of matador documents to filter.

        Returns:
            list: the filtered cursor.
        """

        return filter_cursor_by_chempots(species, cursor)

    def _setup_per_b_fields(self):
        """ Calculate the enthalpy and volume per "B" in A_x B. """
        if self._per_b_done:
            return
        for ind, doc in enumerate(self.cursor):
            nums_b = len(self.species[1:]) * [0]
            for elem in doc['stoichiometry']:
                for chem_pot_ind, chem_pot in enumerate(self.species[1:]):
                    if elem[0] == chem_pot:
                        nums_b[chem_pot_ind] += elem[1]

            num_b = sum(nums_b)
            num_fu = doc['num_fu']
            if num_b == 0:
                self.cursor[ind][self._extensive_energy_key + '_per_b'] = 12345e5
                self.cursor[ind]['cell_volume_per_b'] = 12345e5
            else:
                self.cursor[ind][self._extensive_energy_key + '_per_b'] = doc[self._extensive_energy_key] / (num_b * num_fu)
                if 'cell_volume' in doc:
                    self.cursor[ind]['cell_volume_per_b'] = doc['cell_volume'] / (num_b * num_fu)

        capacities = np.zeros((len(self.cursor)))
        for i, doc in enumerate(self.cursor):
            concs = self.cursor[i]['concentration']
            concs = get_padded_composition(doc['stoichiometry'], self.elements)
            capacities[i] = get_generic_grav_capacity(concs, self.elements)
        set_cursor_from_array(self.cursor, capacities, 'gravimetric_capacity')
        self._per_b_done = True

    def voltage_curve(self, hull_cursor):
        """ Take a computed convex hull and calculate voltages for either binary or ternary
        systems. Sets the self.voltage_data attribute with various fields.

        Parameters:
            hull_cursor (list(dict)): list of structures to include in the voltage curve.

        """
        if not self._non_elemental:
            self._setup_per_b_fields()
        if self._dimension == 2:
            self._calculate_binary_voltage_curve(hull_cursor)
        elif self._dimension == 3:
            self._calculate_ternary_voltage_curve(hull_cursor)
        else:
            raise RuntimeError('Unable to calculate voltage curve for hull of dimension {}'.format(self._dimension))

        self._scale_voltages(self.species[0])

        for profile in self.voltage_data:
            print(profile.voltage_summary(csv=self.args.get('csv', False)))

    def volume_curve(self):
        """ Take stable compositions and volume and calculate
        volume expansion per "B" in AB binary.

        """

        if not self._non_elemental and not self._per_b_done:
            self._setup_per_b_fields()
        if self._dimension == 2:
            self._calculate_binary_volume_curve()
        elif self._dimension == 3:
            # dimension 3 is implemented inside the voltage curve call.
            pass
        else:
            raise NotImplementedError(
                "Volume curves have only been implemented for binary and ternary phase diagrams."
            )

        self.print_volume_summary()

    def print_volume_summary(self):
        """ Prints a volume data summary.

        If self.args['csv'] is True, save the summary to a file.

        """
        for reaction_idx, _ in enumerate(self.volume_data['Q']):
            data_str = '# Reaction {} \n'.format(reaction_idx+1)
            data_str += '# ' + ''.join(get_formula_from_stoich(self.volume_data['endstoichs'][reaction_idx])) + '\n'
            data_str += '# {:>10}\t{:>14} \t{:>14}\n'.format('Q (mAh/g)', 'Volume (A^3)', 'Volume ratio with bulk')
            for idx, _ in enumerate(self.volume_data['electrode_volume'][reaction_idx]):
                data_str += ('{:>10.2f} \t{:14.2f} \t{:14.2f}'
                             .format(self.volume_data['Q'][reaction_idx][idx],
                                     self.volume_data['electrode_volume'][reaction_idx][idx],
                                     self.volume_data['volume_ratio_with_bulk'][reaction_idx][idx]))
                if idx != len(self.volume_data['Q'][reaction_idx]) - 1:
                    data_str += '\n'
            if self.args.get('csv'):
                with open(''.join(self.species) + '_volume.csv', 'w') as f:
                    f.write(data_str)
            print('\nVolume data:')
            print('\n' + data_str)

    def get_hull_distances(self, *args, **kwargs):
        """ Wrapper to PhaseDiagram.get_hull_distances. """
        return self.phase_diagram.get_hull_distances(*args, **kwargs)

    def _calculate_binary_voltage_curve(self, hull_cursor):
        """ Generate binary voltage curve, setting the self.voltage_data
        dictionary.

        Parameters:
            hull_cursor (list(dict)): list of structures to include in the voltage curve.

        """
        mu_enthalpy = get_array_from_cursor(self.chempot_cursor, self.energy_key)
        # take a copy of the cursor so it can be reordered
        _hull_cursor = deepcopy(hull_cursor)
        x = get_num_intercalated(_hull_cursor)
        _x_order = np.argsort(x)
        capacities = np.asarray([
            get_generic_grav_capacity(
                get_padded_composition(doc['stoichiometry'], self.elements),
                self.elements)
            for doc in _hull_cursor
        ])
        set_cursor_from_array(_hull_cursor, x, 'conc_of_active_ion')
        set_cursor_from_array(_hull_cursor, capacities, 'gravimetric_capacity')
        _hull_cursor = sorted(_hull_cursor, key=lambda doc: doc['conc_of_active_ion'])
        capacities = capacities[_x_order]
        x = np.sort(x)
        stable_enthalpy_per_b = get_array_from_cursor(_hull_cursor,
                                                      self._extensive_energy_key + '_per_b')

        x, uniq_idxs = np.unique(x, return_index=True)
        stable_enthalpy_per_b = stable_enthalpy_per_b[uniq_idxs]
        capacities = capacities[uniq_idxs]

        V = np.zeros_like(x)
        V[1:] = - np.diff(stable_enthalpy_per_b) / np.diff(x) + mu_enthalpy[0]
        V[0] = V[1]

        reactions = [((None, get_formula_from_stoich(doc['stoichiometry'])),) for doc in _hull_cursor[1:]]

        # make V, Q and x available for plotting, stripping NaNs to re-add later
        # in the edge case of duplicate chemical potentials
        capacities, unique_caps = np.unique(capacities, return_index=True)
        non_nan = np.argwhere(np.isfinite(capacities))
        V = (np.asarray(V)[unique_caps])[non_nan].flatten().tolist()
        x = (np.asarray(x)[unique_caps])[non_nan].flatten().tolist()
        capacities = capacities[non_nan].flatten().tolist()

        x.append(np.nan)
        capacities.append(np.nan)
        V.append(0.0)

        average_voltage = Electrode.calculate_average_voltage(capacities, V)
        profile = VoltageProfile(
            voltages=V,
            capacities=capacities,
            average_voltage=average_voltage,
            starting_stoichiometry=[[self.species[1], 1.0]],
            reactions=reactions,
            active_ion=self.species[0]
        )
        self.voltage_data.append(profile)

    def _calculate_ternary_voltage_curve(self, hull_cursor):
        """ Calculate tenary voltage curve, setting self.voltage_data.
        First pass written by James Darby, jpd47@cam.ac.uk.

        Parameters:
            hull_cursor (list(dict)): list of structures to include in the voltage curve.

        """

        # construct working array of concentrations and energies
        stoichs = get_array_from_cursor(hull_cursor, 'stoichiometry')

        # do another convex hull on just the known hull points, to allow access to useful indices
        import scipy.spatial

        # construct working array of concentrations and energies
        points = np.hstack((
            get_array_from_cursor(hull_cursor, 'concentration'),
            get_array_from_cursor(hull_cursor, self.energy_key).reshape(len(hull_cursor), 1)
        ))
        stoichs = get_array_from_cursor(hull_cursor, 'stoichiometry')

        # do another convex hull on just the known hull points, to allow access to useful indices
        convex_hull = scipy.spatial.ConvexHull(points)

        endpoints, endstoichs = Electrode._find_starting_materials(convex_hull.points, stoichs)
        print('{} starting point(s) found.'.format(len(endstoichs)))
        for endstoich in endstoichs:
            print(get_formula_from_stoich(endstoich), end=' ')
        print('\n')

        # iterate over possible delithiated phases
        for reaction_ind, endpoint in enumerate(endpoints):
            print(30 * '-')
            print('Reaction {}, {}:'.format(reaction_ind+1, get_formula_from_stoich(endstoichs[reaction_ind])))
            reactions, capacities, voltages, average_voltage, volumes = self._construct_electrode(
                convex_hull, endpoint, endstoichs[reaction_ind], hull_cursor
            )

            profile = VoltageProfile(
                starting_stoichiometry=endstoichs[reaction_ind],
                reactions=reactions,
                capacities=capacities,
                voltages=voltages,
                average_voltage=average_voltage,
                active_ion=self.species[0]
            )
            self.voltage_data.append(profile)

            if not self._non_elemental:
                self.volume_data['x'].append(capacities)
                self.volume_data['Q'].append(capacities)
                self.volume_data['electrode_volume'].append(volumes)
                self.volume_data['endstoichs'].append(endstoichs[reaction_ind])
                self.volume_data['volume_ratio_with_bulk'].append(np.asarray(volumes) / volumes[0])
                self.volume_data['volume_expansion_percentage'].append(((np.asarray(volumes) / volumes[0]) - 1) * 100)
                self.volume_data['hull_distances'].append(np.zeros_like(capacities))

    def _calculate_binary_volume_curve(self):
        """ Take stable compositions and volume and calculate volume
        expansion per "B" in AB binary.

        """
        stable_comp = get_array_from_cursor(self.hull_cursor, 'concentration')
        stable_comp, unique_comp_inds = np.unique(stable_comp, return_index=True)
        for doc in self.hull_cursor:
            if 'cell_volume_per_b' not in doc:
                raise RuntimeError("Document missing key `cell_volume_per_b`: {}".format(doc))
        stable_stoichs = get_array_from_cursor(self.hull_cursor, 'stoichiometry')[unique_comp_inds]
        stable_vol = get_array_from_cursor(self.hull_cursor, 'cell_volume_per_b')[unique_comp_inds]
        stable_cap = get_array_from_cursor(self.hull_cursor, 'gravimetric_capacity')[unique_comp_inds]
        hull_distances = get_array_from_cursor(self.hull_cursor, 'hull_distance')[unique_comp_inds]
        with np.errstate(divide='ignore'):
            stable_x = stable_comp / (1 - stable_comp)
        non_nans = np.argwhere(np.isfinite(stable_x))
        self.volume_data['x'].append(stable_x[non_nans].flatten())
        self.volume_data['Q'].append(stable_cap[non_nans].flatten())
        self.volume_data['electrode_volume'].append(stable_vol[non_nans].flatten())
        self.volume_data['volume_ratio_with_bulk'].append((stable_vol[non_nans] / stable_vol[0]).flatten())
        self.volume_data['volume_expansion_percentage'].append(((stable_vol[non_nans] / stable_vol[0]) - 1) * 100)
        self.volume_data['hull_distances'].append(hull_distances[non_nans].flatten())
        self.volume_data['endstoichs'].append(stable_stoichs[non_nans].flatten()[0])

    def _get_species(self, species=None, elements=None):
        """ Try to determine species for phase diagram from arguments or
        passed data. Sets the `self.species` attribute.

        Keyword arguments:
            species (str/list): the species kwarg passed to __init__.
            elements (str/list): the elements kwarg passed to __init__.

        """
        # try to determine species for phase diagram, from arguments and data
        species = species or elements
        if species is None:
            # handles element list passed by query: should only ever be a list of one entry
            if self.args.get('composition') is not None:
                if isinstance(self.args.get('composition'), list):
                    species = self.args.get('composition')[0]
                else:
                    species = self.args.get('composition')
                if ':' in species:
                    species = species.split(':')
        if species is None:
            if self.from_cursor:
                species = list({atom for doc in self.cursor for atom in doc['atom_types']})
            else:
                raise RuntimeError("Unable to determine species in hull, please specify with "
                                   "e.g. `elements=['Li', 'P']")

        # handles when species is e.g. ['LiCo2:Sn2S']
        if isinstance(species, str):
            if ':' in species:
                species = species.split(':')
            else:
                species = [spec for spec in re.split(r'([A-Z][a-z]*)', species) if spec]

        # edge case where user passes e.g. ['KP'], when they mean ['K', 'P'].
        if isinstance(species, list) and len(species) == 1:
            tmp_species = []
            for spec in species:
                if ':' in spec:
                    tmp_species.append(spec.split(':'))
                else:
                    tmp_species.extend([item for item in re.split(r'([A-Z][a-z]*)', spec) if item])
            species = tmp_species

        return species

    def _construct_electrode(self, hull, endpoint, endstoich, hull_cursor):

        intersections, crossover = self._find_hull_pathway_intersections(
            endpoint, endstoich, hull
        )
        return self._compute_voltages_from_intersections(
            intersections, hull, crossover, endstoich, hull_cursor
        )

    def _compute_voltages_from_intersections(
            self, intersections, hull, crossover, endstoich, hull_cursor
    ):
        """ Traverse the composition pathway and its intersections with hull facets
        and compute the voltage and volume drops along the way, for a ternary system
        with N 3-phase regions.

        Parameters:
            intersections (np.ndarray): Nx3 array containing the list of face indices
                crossed by the path.
            hull (scipy.spatial.ConvexHull): the temporary hull object that is referred
                to by any indices.
            crossover (np.ndarray): (N+1)x3 array containing the coordinates at which
                the composition pathway crosses the hull faces.
            endstoich (list): a matador stoichiometry that defines the end of the pathway.
            hull_cursor (list): the actual structures used to create the hull.

        """
        if len(crossover) != len(intersections) + 1:
            raise RuntimeError("Incompatible number of crossovers ({}) and intersections ({})."
                               .format(np.shape(crossover), np.shape(intersections)))

        # set up output arrays
        voltages = np.empty(len(intersections) + 1)
        volumes = np.empty(len(intersections))
        reactions = []

        # set up some useful arrays for later
        points = hull.points
        mu_enthalpy = get_array_from_cursor(self.chempot_cursor, self.energy_key)
        if not self._non_elemental:
            input_volumes = get_array_from_cursor(hull_cursor, 'cell_volume') / get_array_from_cursor(hull_cursor, 'num_atoms')
        capacities = [get_generic_grav_capacity(point, self.species) for point in crossover]

        # initial composition of one formula unit of the starting material
        initial_comp = get_padded_composition(endstoich, self.species)

        # loop over intersected faces and compute the voltage and volume at the
        # crossover with that face
        for ind, face in enumerate(intersections):

            simplex_index = int(face[0])

            final_stoichs = [hull_cursor[idx]['stoichiometry'] for idx in hull.simplices[simplex_index]]
            atoms_per_fu = [sum([elem[1] for elem in stoich]) for stoich in final_stoichs]

            energy_vec = points[hull.simplices[simplex_index], 2]
            comp = points[hull.simplices[simplex_index], :]
            comp[:, 2] = 1 - comp[:, 0] - comp[:, 1]

            # normalize the crossover composition to one formula unit of the starting electrode
            norm = np.asarray(crossover[ind][1:]) / np.asarray(initial_comp[1:])
            ratios_of_phases = np.linalg.solve(comp.T, crossover[ind] / norm[0])

            # remove small numerical noise
            ratios_of_phases[np.where(ratios_of_phases < EPS)] = 0

            # create a list containing the sections of the balanced reaction for printing
            balanced_reaction = []
            for i, ratio in enumerate(ratios_of_phases):
                if ratio < EPS:
                    continue
                else:
                    formula = get_formula_from_stoich(final_stoichs[i])
                    balanced_reaction.append((ratio / atoms_per_fu[i], formula))
            reactions.append(balanced_reaction)

            # compute the voltage as the difference between the projection of the gradient down to pure active ion,
            # and the reference chemical potential
            comp_inv = np.linalg.inv(comp.T)
            V = -(comp_inv.dot([1, 0, 0])).dot(energy_vec)
            V = V + mu_enthalpy[0]

            # compute the volume of the final electrode
            if not self._non_elemental:
                if ind <= len(intersections) - 1:
                    volume_vec = input_volumes[hull.simplices[simplex_index]]
                    volumes[ind] = np.dot(ratios_of_phases, volume_vec)

            # double up on first voltage
            if ind == 0:
                voltages[0] = V
            voltages[ind+1] = V

        average_voltage = Electrode.calculate_average_voltage(capacities, voltages)

        # print the reaction over a few lines, remove 1s and rounded ratios
        print(
            ' ---> '.join(
                [' + '.join(
                    ["{}{}".format(
                        str(round(chem[0], 3)) + ' ' if abs(chem[0] - 1) > EPS else '',
                        chem[1])
                     for chem in region])
                 for region in reactions])
        )

        return reactions, capacities, voltages, average_voltage, volumes

    def _find_hull_pathway_intersections(self, endpoint, endstoich, hull):
        """ This function traverses a ternary phase diagram on a path from `endpoint`
        towards [1, 0, 0], i.e. a pure phase of the active ion. The aim is to find
        all of the faces and intersections along this pathway, making sure to only
        add unique intersections, and making appropriate (symmetry-breaking) choices
        of which face to use. We proceed as follows:

            1. Filter out any vertices that contain only binary phases, or that contain
                only the chemical potentials.
            2. Loop over all faces of the convex hull and test if the pathway touches/intersects
                that face. If it does, ascertain where the intersection occurs, and whether
                it is at one, two or none of the vertices of the face. Create an array of unique
                intersections for this face and save them for later.
            3. Loop over the zeros found between for each face. If the pathway only grazes a single
                point on this face, ignore it. If one of the edges is parallel to the pathway,
                make sure to include this face and the faces either side of it.

        Parameters:
            endpoint (np.ndarray): array containing composition (2D) and energy of
                start point.

        """
        import scipy.spatial.distance
        import itertools

        endpoint = endpoint[:-1]
        compositions = np.zeros_like(hull.points)
        compositions[:, 0] = hull.points[:, 0]
        compositions[:, 1] = hull.points[:, 1]
        compositions[:, 2] = 1 - hull.points[:, 0] - hull.points[:, 1]

        # define the composition pathway through ternary space
        gradient = endpoint[1] / (endpoint[0] - 1)
        # has to intersect [1, 0] so c = y0 - m*x0 = -m
        y0 = -gradient

        # filter the vertices we're going to consider
        skip_inds = set()
        for simp_ind, simplex in enumerate(hull.simplices):
            vertices = np.asarray([compositions[i] for i in simplex])
            # skip all simplices that contain only binary phases
            for i in range(3):
                if np.max(vertices[:, i]) < BOUNDARY_EPS:
                    skip_inds.add(simp_ind)
                    continue
            # skip the outer triangle formed by the chemical potentials
            if all(np.max(vertices, axis=-1) > 1 - BOUNDARY_EPS):
                skip_inds.add(simp_ind)
                continue

        two_phase_crossover = []
        three_phase_crossover = []

        for simp_ind, simplex in enumerate(hull.simplices):
            if simp_ind in skip_inds:
                continue
            vertices = np.asarray([compositions[i] for i in simplex])

            # put each vertex of triangle into line equation and test their signs
            test = vertices[:, 0] * gradient + y0 - vertices[:, 1]
            test[np.abs(test) < BOUNDARY_EPS] = 0.0
            test = np.sign(test)

            # if there are two different signs in the test array, then the line intersects/grazes this face triangle
            if len(np.unique(test)) > 1:
                # now find intersection points and split them into one, two and three-phase crossover regions
                zeros = [val for zero in np.where(test == 0) for val in zero.tolist()]
                num_zeros = len(zeros)
                zero_points = []
                # skip_vertex = None
                for zero in zeros:
                    zero_pos = compositions[simplex[zero]]
                    if num_zeros == 2:
                        zero_points.append(zero_pos)
                        two_phase_crossover.append((zero_pos, simp_ind))

                # if we have already found both crossovers, skip to next face
                if num_zeros == 2:
                    continue

                # now find the non-trivial intersections
                for i, j in itertools.combinations(simplex, r=2):
                    # if skip_vertex in [i, j]:
                    #     continue

                    A = compositions[i]
                    B = compositions[j]
                    C = endpoint
                    D = [1, 0]

                    def coeffs(X, Y):
                        # find line equation for edge AB
                        # a x + b y = c
                        return (Y[1] - X[1], X[0] - Y[0])

                    a1, b1 = coeffs(A, B)
                    c1 = a1 * A[0] + b1 * A[1]

                    a2, b2 = coeffs(C, D)
                    c2 = a2 * C[0] + b2 * C[1]
                    det = a1 * b2 - a2 * b1
                    if det == 0:
                        continue

                    x = (b2 * c1 - b1 * c2) / det
                    y = (a1 * c2 - a2 * c1) / det
                    intersection_point = np.asarray([x, y, 1-x-y])

                    # now have to test whether that intersection point is inside the triangle
                    # which we can do by testing that is inside the rectangle spanned by AB
                    x_bound = sorted([A[0], B[0]])
                    y_bound = sorted([A[1], B[1]])
                    if (
                        x_bound[0] - BOUNDARY_EPS/2 <= intersection_point[0] <= x_bound[1] + BOUNDARY_EPS/2 and
                        y_bound[0] - BOUNDARY_EPS/2 <= intersection_point[1] <= y_bound[1] + BOUNDARY_EPS/2
                    ):
                        # three_phase_crossover.append((intersection_point, simp_ind))
                        zero_points.append(intersection_point)

                unique_zeros = []
                for zero in zero_points:
                    if (len(unique_zeros) >= 1 and
                            np.any(scipy.spatial.distance.cdist(unique_zeros, zero.reshape(1, 3)) < BOUNDARY_EPS)):
                        pass
                    else:
                        unique_zeros.append(zero)

                num_zeros = len(unique_zeros)
                if num_zeros == 1:
                    # we never need to use faces which just touch the triangle
                    continue
                elif num_zeros == 2:
                    three_phase_crossover.append((unique_zeros[0], simp_ind))
                    three_phase_crossover.append((unique_zeros[1], simp_ind))

        # loop over points and only add the unique ones to the crossover array
        # if a point exists as a two-phase region, then always add that simplex
        crossover = [[1, 0, 0]]
        simplices = {}
        zeros = sorted(two_phase_crossover + three_phase_crossover, key=lambda x: x[0][0])
        # for multiplicity, zeros in enumerate([one_phase_crossover, two_phase_crossover, three_phase_crossover]):
        for zero, simp_ind in zeros:
            if len(crossover) >= 1 and np.any(scipy.spatial.distance.cdist(crossover, zero.reshape(1, 3)) < BOUNDARY_EPS):
                pass
            else:
                if simp_ind not in simplices:
                    crossover.append(zero)
                    # if multiplicity > 0:
                    simplices[simp_ind] = zero[0]

        intersections = sorted(simplices.items(), key=lambda x: x[1])

        if len(intersections) == 0:
            raise RuntimeError(
                'No intermediate structures found for starting point {}.'
                .format(get_formula_from_stoich(endstoich))
            )

        return intersections, sorted(crossover, key=lambda x: x[0])

    def _scale_voltages(self, species):
        """ Rescale voltages to account for the valence of the active ion,
        as stored in the Electrode class.

        """
        valence_factor = Electrode.valence_data.get(species, None)

        if valence_factor is None:
            warnings.warn("Unable to find valence of species {}, treating it as 1.".format(species))
            return

        if valence_factor != 1:
            for ind, start_point in enumerate(self.voltage_data):
                for j in range(len(self.voltage_data[ind].voltages)):
                    self.voltage_data[ind].voltages[j] *= valence_factor
            self.voltage_data[ind].average_voltage *= valence_factor



# coding: utf-8
# Distributed under the terms of the MIT License.

""" This submodule contains functions to plot temperature-dependent
quantities for particular structures, e.g. free energies and heat
capacities.

"""

from matador.plotting.plotting import plotting_function

__all__ = ['plot_free_energy']


@plotting_function
def plot_free_energy(model, temperatures=None, ax=None, line_kwargs=None, **kwargs):
    """ Plot G(T) on the array of given temperatures. Default T is [0, 800].

    Keyword arguments:
        temperatures (list/np.ndarray): list or array of temperatures to plot.
            If the array/list has length 2, use these as the start and endpoints
            with 21 plotting points.
        ax (matplotlib.pyplot.Axis): axis object to plot onto.

    """

    import numpy as np
    import matplotlib.pyplot as plt

    legend = isinstance(line_kwargs, dict) and 'label' in line_kwargs and ax is not None

    if line_kwargs is None:
        line_kwargs = {}

    num_points = 21
    start = 0
    end = 800
    if temperatures is not None and len(temperatures) == 2:
        start = temperatures[0]
        end = temperatures[1]

    if temperatures is None or len(temperatures) == 2:
        _temperatures = np.linspace(start, end, num_points)
    elif temperatures is not None:
        _temperatures = temperatures

    if ax is None:
        fig, ax = plt.subplots()

    t, e = model.vibrational_free_energy(temperatures=_temperatures)
    ax.plot(t, e, **line_kwargs)
    ax.set_ylabel("Vibrational free energy (eV/atom)")
    ax.set_xlabel("Temperature (K)")
    if legend:
        ax.legend()

    return ax



#!/usr/bin/env python

""" Some tests for high-throughput calculations. """

import unittest
import os
from os.path import isfile, isdir
import glob
import copy
import shutil
import time
import warnings
import multiprocessing as mp

import psutil

from matador.utils.errors import (
    CalculationError,
    MaxMemoryEstimateExceeded,
    CriticalError,
    WalltimeError,
    InputError,
)
from matador.compute import ComputeTask, BatchRun, reset_job_folder
from matador.compute.slurm import SlurmQueueManager
from matador.compute.pbs import PBSQueueManager
from matador.scrapers.castep_scrapers import (
    cell2dict,
    param2dict,
    res2dict,
    castep2dict,
)
from .utils import REAL_PATH, MatadorUnitTest, detect_program

HOSTNAME = os.uname()[1]
PATHS_TO_DEL = ["completed", "bad_castep", "input", "logs", HOSTNAME]
VERBOSITY = 10
EXECUTABLE = "castep"
RUN_SLOW_TESTS = HOSTNAME == "cluster2"

CASTEP_PRESENT = detect_program(EXECUTABLE)
MPI_PRESENT = detect_program("mpirun")

if CASTEP_PRESENT and MPI_PRESENT:
    NCORES = max(psutil.cpu_count(logical=False) - 2, 1)
else:
    NCORES = 1


class ComputeTest(MatadorUnitTest):
    """ Run tests equivalent to using the run3 script for
    various artificial setups.

    """

    def test_missing_exec(self):
        """ Ensure failure if exec misses. """
        cell_dict, s = cell2dict(
            REAL_PATH + "/data/LiAs_tests/LiAs.cell", verbosity=VERBOSITY, db=False
        )
        assert s
        param_dict, s = param2dict(
            REAL_PATH + "/data/LiAs_tests/LiAs.param", verbosity=VERBOSITY, db=False
        )
        assert s

        node = None
        nnodes = None
        seed = REAL_PATH + "/data/structures/LiAs_testcase.res"

        fall_over = False

        try:
            ComputeTask(
                ncores=NCORES,
                nnodes=nnodes,
                node=node,
                res=seed,
                param_dict=param_dict,
                cell_dict=cell_dict,
                verbosity=VERBOSITY,
                killcheck=True,
                reopt=False,
                executable="THIS WAS MEANT TO FAIL, DON'T WORRY",
                start=True,
            )
        except CriticalError:
            fall_over = True

        self.assertTrue(fall_over)

    def test_file_not_written(self):
        """ Run a calculation with an executable that only does "sleep" and
        check that run3 will stop the calculation early as no file is written.

        """
        seed = REAL_PATH + "data/symmetry_failure/Sb.res"
        cell_dict, s = cell2dict(
            REAL_PATH + "/data/symmetry_failure/KSb.cell", verbosity=VERBOSITY, db=False
        )
        assert s
        param_dict, s = param2dict(
            REAL_PATH + "/data/symmetry_failure/KSb.param",
            verbosity=VERBOSITY,
            db=False,
        )
        assert s
        executable = REAL_PATH + "data/missing_file_test/monkey_patch_sleep.sh"
        node = None

        relaxer = ComputeTask(
            ncores=NCORES,
            nnodes=None,
            node=node,
            res=seed,
            param_dict=param_dict,
            cell_dict=cell_dict,
            verbosity=VERBOSITY,
            executable=executable,
            exec_test=False,
            polltime=1,
            start=False,
        )
        with self.assertRaises(CalculationError):
            relaxer.begin()

        self.assertTrue(relaxer.final_result is None)

    def test_old_file(self):
        """ Run a calculation with an executable that only does "sleep", in the
        presence of a file that was written previouisly, and check that run3
        will stop the calculation early as no file is written.

        """
        seed = REAL_PATH + "data/symmetry_failure/Sb.res"
        with open("Sb.castep", "w") as f:
            f.write("I am a CASTEP file, for sure.")

        cell_dict, s = cell2dict(
            REAL_PATH + "/data/symmetry_failure/KSb.cell", verbosity=VERBOSITY, db=False
        )
        assert s
        param_dict, s = param2dict(
            REAL_PATH + "/data/symmetry_failure/KSb.param",
            verbosity=VERBOSITY,
            db=False,
        )
        assert s
        executable = REAL_PATH + "data/missing_file_test/monkey_patch_sleep.sh"
        node = None

        relaxer = ComputeTask(
            ncores=NCORES,
            nnodes=None,
            node=node,
            res=seed,
            param_dict=param_dict,
            cell_dict=cell_dict,
            verbosity=VERBOSITY,
            executable=executable,
            exec_test=False,
            polltime=1,
            start=False,
        )

        with self.assertRaises(CalculationError):
            relaxer.begin()
        self.assertTrue(relaxer.final_result is None)

    def test_faked_error_recovery(self):
        """ Run a calculation that *should* throw a symmetry error, and try to
        recover from the error. If CASTEP is not present, monkey patch such that
        ComputeTask copies the output files it would have expected.

        """
        seed = REAL_PATH + "data/symmetry_failure/Sb.res"
        cell_dict, s = cell2dict(
            REAL_PATH + "/data/symmetry_failure/KSb.cell", verbosity=VERBOSITY, db=False
        )
        assert s
        param_dict, s = param2dict(
            REAL_PATH + "/data/symmetry_failure/KSb.param",
            verbosity=VERBOSITY,
            db=False,
        )
        assert s
        ncores = 1
        executable = REAL_PATH + "data/symmetry_failure/monkey_patch_move.sh"
        node = None

        relaxer = ComputeTask(
            ncores=ncores,
            nnodes=None,
            node=node,
            res=seed,
            param_dict=param_dict,
            cell_dict=cell_dict,
            verbosity=VERBOSITY,
            executable=executable,
            exec_test=False,
            polltime=1,
            start=False,
        )

        with self.assertRaises(CalculationError):
            relaxer.begin()

        bad_castep_exists = isdir("bad_castep")
        completed_exists = isdir("completed")

        self.assertTrue(bad_castep_exists)
        self.assertFalse(completed_exists)
        self.assertTrue(relaxer.final_result is None)
        self.assertEqual(relaxer._num_retries, 3)
        self.assertTrue("symmetry_generate" not in relaxer.calc_doc)
        self.assertTrue("snap_to_symmetry" not in relaxer.calc_doc)
        self.assertTrue("symmetry_tol" not in relaxer.calc_doc)

    @unittest.skipIf(
        (not CASTEP_PRESENT or not MPI_PRESENT),
        "castep or mpirun executable not found in PATH",
    )
    def test_relax_to_queue(self):
        """ Mimic GA and test Queue relaxations. """

        newborn, s = res2dict(
            REAL_PATH + "/data/structures/LiAs_testcase.res",
            verbosity=VERBOSITY,
            db=False,
        )
        assert s
        cell_dict, s = cell2dict(
            REAL_PATH + "/data/LiAs_tests/LiAs.cell", verbosity=VERBOSITY, db=False
        )
        assert s
        param_dict, s = param2dict(
            REAL_PATH + "/data/LiAs_tests/LiAs.param", verbosity=VERBOSITY, db=False
        )
        assert s

        node = None
        executable = "castep"
        newborn["source"] = [REAL_PATH + "/data/GA_TESTCASE.res"]

        shutil.copy(REAL_PATH + "data/pspots/Li_00PBE.usp", ".")
        shutil.copy(REAL_PATH + "data/pspots/As_00PBE.usp", ".")

        queue = mp.Queue()
        relaxer = ComputeTask(
            ncores=NCORES,
            nnodes=None,
            node=node,
            res=newborn,
            param_dict=param_dict,
            cell_dict=cell_dict,
            verbosity=VERBOSITY,
            killcheck=True,
            reopt=False,
            executable=executable,
            output_queue=queue,
            start=True,
        )
        # store proc object with structure ID, node name, output queue and number of cores
        proc = (1, node, mp.Process(target=relaxer.relax), NCORES)
        proc[2].start()
        while proc[2].is_alive():
            time.sleep(1)

        result, success = castep2dict("completed/GA_TESTCASE.castep")
        queue_result = queue.get()

        match_dict = dict()
        for key in queue_result:
            if key in ["source", "site_occupancy", "geom_iter"]:
                continue
            match_dict[key] = queue_result[key] == result[key]
            if not match_dict[key]:
                print(key, queue_result[key], result[key])

        completed_exists = isfile("completed/GA_TESTCASE.res")
        input_exists = isfile("input/GA_TESTCASE.res")

        self.assertTrue(completed_exists, "couldn't find output file!")
        self.assertTrue(input_exists, "couldn't find shutil.copy of input file!")
        self.assertTrue(success, "couldn't parse output file!")
        self.assertTrue(all([match_dict[key] for key in match_dict]))

    @unittest.skipIf(
        (not CASTEP_PRESENT or not MPI_PRESENT),
        "castep or mpirun executable not found in PATH",
    )
    def test_relax_to_file(self):
        """ Relax structure from file to file. """
        seed = "_Li.res"
        shutil.copy(REAL_PATH + "data/structures/Li.res", "_Li.res")

        cell_dict, s = cell2dict(
            REAL_PATH + "/data/LiAs_tests/LiAs.cell", verbosity=VERBOSITY, db=False
        )
        assert s
        param_dict, s = param2dict(
            REAL_PATH + "/data/LiAs_tests/LiAs.param", verbosity=VERBOSITY, db=False
        )
        assert s
        executable = "castep"
        node = None

        shutil.copy(REAL_PATH + "data/pspots/Li_00PBE.usp", ".")
        shutil.copy(REAL_PATH + "data/pspots/As_00PBE.usp", ".")

        ComputeTask(
            ncores=NCORES,
            nnodes=None,
            node=node,
            res=seed,
            param_dict=param_dict,
            cell_dict=cell_dict,
            verbosity=VERBOSITY,
            killcheck=True,
            reopt=True,
            executable=executable,
            exec_test=False,
            start=True,
        )

        print("Process completed!")

        completed_exists = isfile("completed/_Li.res")
        base_files_exist = [
            isfile("_Li.res"),
            isfile("_Li.res.lock"),
            isfile("_Li.castep"),
        ]
        self.assertTrue(completed_exists, "couldn't find output file!")
        self.assertFalse(any(base_files_exist), "couldn't clean input files")

    @unittest.skipIf(
        (not CASTEP_PRESENT or not MPI_PRESENT),
        "castep or mpirun executable not found in PATH",
    )
    def test_failed_relaxation(self):
        """ Set a relaxation up to fail. """
        seed = "_LiAs_testcase.res"
        shutil.copy(
            REAL_PATH + "data/structures/LiAs_testcase_bad.res", "_LiAs_testcase.res"
        )

        cell_dict, s = cell2dict(
            REAL_PATH + "/data/LiAs_tests/LiAs.cell", verbosity=VERBOSITY, db=False
        )
        assert s
        param_dict, s = param2dict(
            REAL_PATH + "/data/LiAs_tests/LiAs.param", verbosity=VERBOSITY, db=False
        )
        assert s
        param_dict["geom_max_iter"] = 3
        executable = "castep"
        node = None

        shutil.copy(REAL_PATH + "data/pspots/Li_00PBE.usp", ".")
        shutil.copy(REAL_PATH + "data/pspots/As_00PBE.usp", ".")

        relaxer = ComputeTask(
            ncores=NCORES,
            nnodes=None,
            node=node,
            res=seed,
            param_dict=param_dict,
            cell_dict=cell_dict,
            verbosity=VERBOSITY,
            killcheck=True,
            memcheck=False,
            reopt=True,
            executable=executable,
            rough=0,
            fine_iter=3,
            start=False,
        )

        with self.assertRaises(CalculationError):
            relaxer.begin()

        bad_exists = isfile("bad_castep/_LiAs_testcase.res")

        num = reset_job_folder()

        self.assertTrue(bad_exists, "couldn't find output file!")
        self.assertEqual(num, 0)

    def test_dont_restart_completed_calc(self):
        """ Set a relaxation up to fail. """

        shutil.copy(
            REAL_PATH
            + "data/no_steps_left_todo/cache/NaP_intermediates_stopped_early.res",
            ".",
        )
        shutil.copy(
            REAL_PATH
            + "data/no_steps_left_todo/cache/NaP_intermediates_stopped_early.castep",
            ".",
        )

        cell_dict, s = cell2dict(
            REAL_PATH + "data/no_steps_left_todo/NaP.cell",
            verbosity=VERBOSITY,
            db=False,
        )
        self.assertTrue(s)
        param_dict, s = param2dict(
            REAL_PATH + "data/no_steps_left_todo/NaP.param",
            verbosity=VERBOSITY,
            db=False,
        )
        self.assertTrue(s)
        executable = "castep"
        node = None
        seed = "NaP_intermediates_stopped_early.res"
        with self.assertRaises(CalculationError):
            ComputeTask(
                ncores=NCORES,
                nnodes=None,
                node=node,
                res=seed,
                param_dict=param_dict,
                cell_dict=cell_dict,
                verbosity=VERBOSITY,
                killcheck=True,
                memcheck=False,
                exec_test=False,
                reopt=True,
                executable=executable,
                max_walltime=5,
                start=True,
            )

        print("Process completed!")

        bad_exists = []
        bad_exists.append(isfile("bad_castep/NaP_intermediates_stopped_early.res"))
        bad_exists.append(isfile("bad_castep/NaP_intermediates_stopped_early.castep"))
        bad_exists = all(bad_exists)
        good_exists = all(isdir(path) for path in ["input", "bad_castep", "logs"])

        self.assertTrue(bad_exists)
        self.assertTrue(good_exists)

    @unittest.skipIf(
        (not CASTEP_PRESENT or not MPI_PRESENT),
        "castep or mpirun executable not found in PATH",
    )
    def test_memcheck(self):
        """ Test the memory checker will not proceed with huge jobs. """

        shutil.copy(
            REAL_PATH + "data/structures/LiAs_testcase.res", "_LiAs_testcase.res"
        )
        shutil.copy(REAL_PATH + "data/pspots/Li_00PBE.usp", ".")
        shutil.copy(REAL_PATH + "data/pspots/As_00PBE.usp", ".")

        cell_dict, s = cell2dict(
            REAL_PATH + "data/LiAs_tests/LiAs.cell", verbosity=VERBOSITY, db=False
        )
        self.assertTrue(s)
        param_dict, s = param2dict(
            REAL_PATH + "data/LiAs_tests/LiAs.param", verbosity=VERBOSITY, db=False
        )
        self.assertTrue(s)

        with self.assertRaises(MaxMemoryEstimateExceeded):
            ComputeTask(
                ncores=NCORES,
                nnodes=None,
                node=None,
                res="_LiAs_testcase.res",
                param_dict=param_dict,
                cell_dict=cell_dict,
                verbosity=VERBOSITY,
                killcheck=True,
                memcheck=True,
                maxmem=1,
                start=True,
            )

        files_that_should_not_exist = ["_LiAs_testcase.res.lock", "jobs.txt"]
        folders_that_should_exist = ["logs"]
        folders_that_should_not_exist = ["bad_castep", "input", "completed"]

        correct_files = all(
            [not isfile(_file) for _file in files_that_should_not_exist]
        )
        correct_folders = all([isdir(folder) for folder in folders_that_should_exist])
        correct_folders *= all(
            [not isdir(folder) for folder in folders_that_should_not_exist]
        )

        self.assertTrue(correct_folders)
        self.assertTrue(correct_files)

    @unittest.skipIf(
        (not CASTEP_PRESENT or not MPI_PRESENT),
        "castep or mpirun executable not found in PATH",
    )
    @unittest.skipIf(not RUN_SLOW_TESTS, "this is a slow test, skipping")
    def test_batch_relax(self):
        """ Batch relax structures from file to file. """

        shutil.copy(REAL_PATH + "data/structures/LiC.res", "_LiC.res")
        shutil.copy(REAL_PATH + "data/LiC_tests/LiC.cell", ".")
        shutil.copy(REAL_PATH + "data/LiC_tests/LiC.param", ".")

        shutil.copy(REAL_PATH + "data/pspots/Li_00PBE.usp", ".")
        shutil.copy(REAL_PATH + "data/pspots/C_00PBE.usp", ".")

        runner = BatchRun(
            seed=["LiC"],
            debug=False,
            no_reopt=True,
            verbosity=VERBOSITY,
            ncores=NCORES,
            executable=EXECUTABLE,
        )
        runner.spawn(join=False)

        completed_exists = isfile("completed/_LiC.res")
        cruft = glob.glob("_LiC*")
        cruft_doesnt_exist = bool(len(cruft))

        num = reset_job_folder()

        self.assertEqual(num, 0)
        self.assertTrue(completed_exists, "couldn't find output file!")
        self.assertFalse(cruft_doesnt_exist, "found some cruft {}".format(cruft))

    def test_batch_queues(self):
        """ Test the scraping of queuing environments. """

        shutil.copy(REAL_PATH + "data/structures/LiC.res", "_LiC.res")
        shutil.copy(REAL_PATH + "data/LiC_tests/LiC.cell", ".")
        shutil.copy(REAL_PATH + "data/LiC_tests/LiC.param", ".")

        shutil.copy(REAL_PATH + "data/pspots/Li_00PBE.usp", ".")
        shutil.copy(REAL_PATH + "data/pspots/C_00PBE.usp", ".")

        slurm_env = {
            "SLURM_NTASKS": "120",
            "SLURM_JOB_ID": "123456",
            "SLURM_ARRAY_TASK_ID": "123123",
            "SLURM_MEM_PER_CPU": "1024",
            "SLURM_RANDOM_STRING": "hello",
            "blah": "abc",
        }

        pbs_env = {"PBS_TASKNUM": "120", "PBS_JOB_ID": "999", "PBS_ARRAYID": "123123"}

        old_env = copy.deepcopy(os.environ)
        os.environ.update(slurm_env)

        runner = BatchRun(
            seed=["LiC"],
            debug=False,
            ncores=NCORES,
            verbosity=VERBOSITY,
            executable=EXECUTABLE,
        )

        self.assertEqual(runner.args["ncores"], NCORES)
        self.assertEqual(type(runner.queue_mgr), SlurmQueueManager)
        self.assertEqual(runner.maxmem, 1024 * 120)
        self.assertEqual(runner.queue_mgr.max_memory, 1024 * 120)
        self.assertEqual(runner.queue_mgr.array_id, 123123)
        self.assertEqual(runner.queue_mgr.env["SLURM_RANDOM_STRING"], "hello")
        self.assertTrue("blah" not in runner.queue_mgr.env)

        runner = BatchRun(
            seed=["LiC"],
            debug=False,
            ncores=None,
            verbosity=VERBOSITY,
            executable=EXECUTABLE,
        )

        self.assertEqual(runner.args["ncores"], 120)
        self.assertEqual(type(runner.queue_mgr), SlurmQueueManager)
        self.assertEqual(runner.maxmem, 1024 * 120)
        self.assertEqual(runner.queue_mgr.max_memory, 1024 * 120)
        self.assertEqual(runner.queue_mgr.array_id, 123123)

        os.environ = copy.deepcopy(old_env)
        os.environ.update(pbs_env)

        runner = BatchRun(
            seed=["LiC"],
            debug=False,
            ncores=None,
            verbosity=VERBOSITY,
            executable=EXECUTABLE,
        )

        print(runner.queue_mgr)

        self.assertEqual(runner.args["ncores"], 120)
        self.assertEqual(type(runner.queue_mgr), PBSQueueManager)
        self.assertEqual(runner.maxmem, None)
        self.assertEqual(runner.queue_mgr.max_memory, None)
        self.assertEqual(runner.queue_mgr.array_id, 123123)

        os.environ.update(slurm_env)
        with self.assertRaises(RuntimeError):
            runner = BatchRun(
                seed=["LiC"],
                debug=False,
                ncores=None,
                verbosity=VERBOSITY,
                executable=EXECUTABLE,
            )

        os.environ = copy.deepcopy(old_env)

    @unittest.skipIf(
        (not CASTEP_PRESENT or not MPI_PRESENT),
        "castep or mpirun executable not found in PATH",
    )
    def test_scf(self):
        """ Perform SCF on structure from file. """
        seed = "_LiC.res"
        shutil.copy(REAL_PATH + "data/structures/LiC.res", "_LiC.res")

        cell_dict, s = cell2dict(
            REAL_PATH + "/data/LiC_tests/LiC_scf.cell", verbosity=VERBOSITY, db=False
        )
        assert s
        param_dict, s = param2dict(
            REAL_PATH + "/data/LiC_tests/LiC_scf.param", verbosity=VERBOSITY, db=False
        )
        assert s
        executable = "castep"
        node = None

        shutil.copy(REAL_PATH + "data/pspots/Li_00PBE.usp", ".")
        shutil.copy(REAL_PATH + "data/pspots/C_00PBE.usp", ".")

        ComputeTask(
            ncores=NCORES,
            nnodes=None,
            node=node,
            res=seed,
            param_dict=param_dict,
            cell_dict=cell_dict,
            verbosity=VERBOSITY,
            killcheck=True,
            reopt=True,
            executable=executable,
            compute_dir="/tmp/compute_test",
            start=True,
        )

        completed_exists = [
            isfile("completed/_LiC.res"),
            isfile("completed/_LiC.castep"),
            isfile("completed/_LiC-out.cell"),
        ]
        base_file_exists = [
            isfile("_LiC.res"),
            isfile("_LiC.castep"),
            isfile("_LiC.res.lock"),
        ]

        self.assertFalse(any(base_file_exists), "failed to clean up files!")
        self.assertTrue(all(completed_exists), "couldn't find output files!")

    @unittest.skipIf(
        (not CASTEP_PRESENT or not MPI_PRESENT),
        "castep or mpirun executable not found in PATH",
    )
    def test_scf_max_walltime(self):
        """ Perform SCF on structure from file. """
        seed = "_LiC.res"
        shutil.copy(REAL_PATH + "data/structures/LiC.res", "_LiC.res")

        cell_dict, s = cell2dict(
            REAL_PATH + "/data/LiC_tests/LiC_scf.cell", verbosity=VERBOSITY, db=False
        )
        assert s
        param_dict, s = param2dict(
            REAL_PATH + "/data/LiC_tests/LiC_scf.param", verbosity=VERBOSITY, db=False
        )
        assert s
        executable = "castep"
        node = None

        shutil.copy(REAL_PATH + "data/pspots/Li_00PBE.usp", ".")
        shutil.copy(REAL_PATH + "data/pspots/C_00PBE.usp", ".")

        with self.assertRaises(WalltimeError):
            ComputeTask(
                ncores=NCORES,
                nnodes=None,
                node=node,
                res=seed,
                param_dict=param_dict,
                cell_dict=cell_dict,
                verbosity=VERBOSITY,
                timings=(5, time.time()),
                polltime=2,
                executable=executable,
                compute_dir="/tmp/compute_test",
                start=True,
            )

        base_file_exists = [
            isfile("_LiC.res"),
            isfile("_LiC.castep"),
        ]

        self.assertFalse(isfile("_LiC.res.lock"), "failed to clean up lock")
        self.assertTrue(all(base_file_exists), "failed to keep valid files!")

    @unittest.skipIf(
        (not CASTEP_PRESENT or not MPI_PRESENT),
        "castep or mpirun executable not found in PATH",
    )
    def test_convergence_runner(self):
        """ Check that convergence tests run to completion. """
        shutil.copy(REAL_PATH + "data/structures/Li.res", "_LiAs_testcase.res")
        shutil.copy(REAL_PATH + "data/LiAs_tests/LiAs_scf.cell", ".")
        shutil.copy(REAL_PATH + "data/LiAs_tests/LiAs_scf.param", ".")
        shutil.copy(REAL_PATH + "data/pspots/Li_00PBE.usp", ".")
        shutil.copy(REAL_PATH + "data/pspots/As_00PBE.usp", ".")

        with open("kpt.conv", "w") as f:
            f.write("0.08\n")
            f.write("0.07")

        with open("cutoff.conv", "w") as f:
            f.write("300\n")
            f.write("400")

        runner = BatchRun(
            seed=["LiAs_scf"],
            debug=False,
            conv_cutoff=True,
            conv_kpt=True,
            verbosity=VERBOSITY,
            ncores=NCORES,
            nprocesses=2,
            executable=EXECUTABLE,
        )
        runner.spawn(join=False)

        dirs_exist = [isdir(_dir) for _dir in ["completed_kpts", "completed_cutoff"]]
        bad_castep_exist = isdir("bad_castep")

        files_that_should_not_exist = [
            "_LiAs_testcase_300eV.res",
            "_LiAs_testcase_400eV.res",
            "_LiAs_testcase_0.08A.res",
            "_LiAs_testcase_0.07A.res",
        ]
        do_bad_files_exist = [isfile(_file) for _file in files_that_should_not_exist]

        results = [
            "completed_cutoff/_LiAs_testcase_{}eV.castep".format(cutoff)
            for cutoff in [300, 400]
        ]
        results += [
            "completed_kpts/_LiAs_testcase_{}A.castep".format(kpts)
            for kpts in [0.08, 0.07]
        ]
        files_exist = [isfile(_file) for _file in results]

        self.assertTrue(all(dirs_exist))
        self.assertFalse(bad_castep_exist)
        self.assertTrue(all(files_exist))
        self.assertFalse(any(do_bad_files_exist))

    @unittest.skipIf(
        (not CASTEP_PRESENT or not MPI_PRESENT),
        "castep or mpirun executable not found in PATH",
    )
    def test_batch_failed_scf(self):
        """ Check that SCF failures don't kill everything... """

        shutil.copy(REAL_PATH + "data/structures/Li.res", "_Li.res")
        shutil.copy(REAL_PATH + "data/structures/LiC.res", "_LiC.res")
        shutil.copy(REAL_PATH + "data/pspots/Li_00PBE.usp", ".")
        shutil.copy(REAL_PATH + "data/pspots/C_00PBE.usp", ".")
        shutil.copy(REAL_PATH + "data/fail_scf/LiC_scf.cell", ".")
        shutil.copy(REAL_PATH + "data/fail_scf/LiC_scf.param", ".")

        runner = BatchRun(
            seed=["LiC_scf"],
            debug=False,
            no_reopt=True,
            verbosity=VERBOSITY,
            ncores=NCORES,
            executable=EXECUTABLE,
        )
        runner.spawn()

        completed_folder_exists = isdir("completed")
        bad_castep_folder_exists = isdir("bad_castep")

        seeds = ["_Li.res", "_LiC.res"]
        output_files_exist = all(
            [isfile("bad_castep/{}".format(seed)) for seed in seeds]
        )

        cruft = glob.glob("_Li*")
        cruft_doesnt_exist = bool(len(cruft))

        compute_dir_exist = isdir(HOSTNAME)

        num = reset_job_folder()

        self.assertEqual(num, 0)
        self.assertFalse(completed_folder_exists, "couldn't find output file!")
        self.assertTrue(bad_castep_folder_exists, "couldn't find bad_castep")
        self.assertTrue(output_files_exist, "couldn't find both outputs")
        self.assertFalse(cruft_doesnt_exist, "found some cruft {}".format(cruft))
        self.assertFalse(compute_dir_exist, "Compute dir not cleaned up!")

    def test_failed_compute_dir_scf(self):
        """ Check that using a garbage path for compute dir causes a safe crash. """
        shutil.copy(REAL_PATH + "data/structures/Li.res", "_Li.res")
        shutil.copy(REAL_PATH + "data/structures/LiC.res", "_LiC.res")
        shutil.copy(REAL_PATH + "data/pspots/Li_00PBE.usp", ".")
        shutil.copy(REAL_PATH + "data/pspots/C_00PBE.usp", ".")
        shutil.copy(REAL_PATH + "data/fail_scf/LiC_scf.cell", ".")
        shutil.copy(REAL_PATH + "data/fail_scf/LiC_scf.param", ".")

        runner = BatchRun(
            seed=["LiC_scf"],
            debug=False,
            no_reopt=True,
            scratch_prefix="/this/drive/doesnt/exist",
            verbosity=VERBOSITY,
            ncores=NCORES,
            executable=EXECUTABLE,
        )
        with self.assertRaises(CriticalError):
            runner.spawn()

    @unittest.skipIf(
        (not CASTEP_PRESENT or not MPI_PRESENT),
        "castep or mpirun executable not found in PATH",
    )
    def test_batch_max_walltime_threaded(self):
        """ Check that WallTimeErrors do kill everything... """

        shutil.copy(REAL_PATH + "data/structures/LiAs_testcase.res", ".")
        shutil.copy(REAL_PATH + "data/structures/LiAs_testcase_bad.res", ".")
        shutil.copy(REAL_PATH + "data/pspots/Li_00PBE.usp", ".")
        shutil.copy(REAL_PATH + "data/pspots/As_00PBE.usp", ".")
        shutil.copy(REAL_PATH + "data/max_walltime/LiAs.cell", ".")
        shutil.copy(REAL_PATH + "data/max_walltime/LiAs.param", ".")

        runner = BatchRun(
            seed=["LiAs"],
            debug=False,
            no_reopt=True,
            verbosity=VERBOSITY,
            ncores=2,
            nprocesses=2,
            executable=EXECUTABLE,
            max_walltime=15,
            polltime=10,
        )
        with self.assertRaises(WalltimeError):
            runner.spawn()

        castep_exists = isfile("LiAs_testcase.castep")
        bad_castep_exists = isfile("LiAs_testcase_bad.castep")
        res_exists = isfile("LiAs_testcase.res") and isfile("LiAs_testcase_bad.res")
        lock_exists = isfile("LiAs_testcase.res.lock") or isfile(
            "LiAs_testcase_bad.res.lock"
        )
        compute_dir_exist = isdir(HOSTNAME)

        self.assertTrue(castep_exists, "Could not find castep file!")
        self.assertTrue(bad_castep_exists, "Could not find bad castep file!")
        self.assertTrue(res_exists, "Could not find res file!")
        self.assertFalse(compute_dir_exist, "Compute dir not cleaned up!")
        self.assertFalse(lock_exists, "Lock file was not deleted!")

    def test_generic_batch(self):
        """ Run a calculation in generic mode, check that res files
        are cycled over.

        """

        executable = "echo"
        files = glob.glob(REAL_PATH + "/data/structures/*.res")
        for _file in files:
            shutil.copy(_file, ".")

        runner = BatchRun(
            seed="*.res",
            debug=False,
            mode="generic",
            verbosity=4,
            ncores=1,
            executable=executable,
        )

        runner.spawn(join=True)

        completed_files_exist = all(
            [isfile("completed/" + _file.split("/")[-1]) for _file in files]
        )
        txt_files_exist = all(
            [isfile(_file) for _file in ["jobs.txt", "finished_cleanly.txt"]]
        )
        dirs = ["completed", "input", "logs"]
        dirs_exist = all([isdir(_dir) for _dir in dirs])

        logs = glob.glob("logs/*.log")
        num_logs = len(logs)
        log_lines = []
        for log in logs:
            with open(log, "r") as f:
                log_lines.append(len(f.readlines()))

        self.assertEqual(num_logs, len(files), msg="Not enough log files!")
        self.assertTrue(
            all(lines > 5 for lines in log_lines), msg="Log files were too short!"
        )
        self.assertTrue(completed_files_exist)
        self.assertTrue(dirs_exist)
        self.assertTrue(txt_files_exist)

    def test_generic_batch_threaded(self):
        """ Run a calculation in generic mode, check that res files
        are cycled over and not repeated by multiple threads. Multiple
        threads competing for the same job.

        """

        executable = "echo"
        files = glob.glob(REAL_PATH + "/data/structures/*.res")[0:2]
        for _file in files:
            shutil.copy(_file, ".")

        runner = BatchRun(
            seed="*.res",
            debug=False,
            mode="generic",
            verbosity=4,
            ncores=1,
            nprocesses=2,
            executable=executable,
        )

        runner.spawn(join=True)

        completed_files_exist = isfile("completed/" + _file.split("/")[-1])
        txt_files_exist = all(
            [isfile(_file) for _file in ["jobs.txt", "finished_cleanly.txt"]]
        )
        dirs = ["completed", "input", "logs"]
        dirs_exist = all([isdir(_dir) for _dir in dirs])

        logs = glob.glob("logs/*.log")
        num_logs = len(logs)
        log_lines = []
        for log in logs:
            with open(log, "r") as f:
                log_lines.append(len(f.readlines()))

        with open("jobs.txt", "r") as f:
            jobs_len = len(f.readlines())

        self.assertEqual(num_logs, len(files), msg="Not enough log files!")
        for f in files:
            self.assertTrue(os.path.isfile("input/{}".format(f.split("/")[-1])))
        self.assertTrue(
            all(lines > 5 for lines in log_lines), msg="Log files were too short!"
        )
        self.assertEqual(jobs_len, len(files))
        self.assertTrue(dirs_exist)
        self.assertTrue(completed_files_exist)
        self.assertTrue(txt_files_exist)

    def test_batch_nothing_todo(self):
        """ Check that nothing is done when there's nothing to do... """
        for file in glob.glob(REAL_PATH + "data/nothing_to_do/*.*"):
            shutil.copy(file, ".")

        runner = BatchRun(
            seed=["LiAs"],
            debug=False,
            no_reopt=True,
            verbosity=VERBOSITY,
            ncores=2,
            nprocesses=2,
            executable=EXECUTABLE,
        )
        start = time.time()
        runner.spawn()
        elapsed = time.time() - start
        self.assertTrue(elapsed < 10, "Sluggish to quit!")

    def test_res_name_collision(self):
        """ Check that run3 safely falls over if there is a file called <seed>.res. """
        shutil.copy(REAL_PATH + "data/file_collision/LiAs.cell", ".")
        shutil.copy(REAL_PATH + "data/file_collision/LiAs.res", ".")
        shutil.copy(REAL_PATH + "data/file_collision/LiAs.param", ".")

        failed_safely = False
        try:
            runner = BatchRun(
                seed=["LiAs"],
                debug=False,
                no_reopt=True,
                verbosity=VERBOSITY,
                ncores=2,
                nprocesses=2,
                executable=EXECUTABLE,
            )
            runner.spawn()
        except InputError:
            failed_safely = True

        self.assertTrue(failed_safely)

    def test_missing_basics(self):
        """" Check that run3 falls over when e.g. xc_functional is missing. """
        for file in glob.glob(REAL_PATH + "data/misisng_basics/*.*"):
            shutil.copy(file, ".")

        tests = ["missing_cutoff", "missing_kpts", "missing_xc", "missing_pspot"]
        errors = [False for test in tests]
        for ind, test in enumerate(tests):
            try:
                runner = BatchRun(
                    seed=["LiC_" + test],
                    debug=False,
                    no_reopt=True,
                    verbosity=VERBOSITY,
                    executable=EXECUTABLE,
                )
                runner.spawn()
            except InputError:
                errors[ind] = True

        self.assertTrue(all(errors))


class BenchmarkCastep(MatadorUnitTest):
    """ Run some short CASTEP calculations and compare the timings
    to single core & multicore references.

    """

    @unittest.skipIf(
        (not CASTEP_PRESENT or not MPI_PRESENT),
        "castep or mpirun executable not found in PATH",
    )
    def test_benchmark_dual_core_scf(self):
        """ Test the time taken to perform a set number of SCF steps
        on 2 cores. CASTEP prints no total timing data for single core jobs.

        """
        from os import makedirs

        seed = "_LiC.res"
        shutil.copy(REAL_PATH + "data/structures/LiC.res", "_LiC.res")

        cell_dict, s = cell2dict(
            REAL_PATH + "/data/benchmark/LiC_scf/LiC_scf.cell",
            verbosity=VERBOSITY,
            db=False,
        )
        self.assertTrue(s)
        param_dict, s = param2dict(
            REAL_PATH + "/data/benchmark/LiC_scf/LiC_scf.param",
            verbosity=VERBOSITY,
            db=False,
        )
        self.assertTrue(s)

        shutil.copy(REAL_PATH + "data/pspots/Li_00PBE.usp", ".")
        shutil.copy(REAL_PATH + "data/pspots/C_00PBE.usp", ".")
        with self.assertRaises(CalculationError):
            ComputeTask(
                ncores=2,
                nnodes=None,
                node=None,
                res=seed,
                param_dict=param_dict,
                cell_dict=cell_dict,
                verbosity=0,
                executable=EXECUTABLE,
                start=True,
            )

        outputs_exist = [
            isfile("bad_castep/_LiC.res"),
            isfile("bad_castep/_LiC.castep"),
        ]

        results, s = castep2dict("bad_castep/_LiC.castep", db=False)
        makedirs(REAL_PATH + "/data/benchmark/results", exist_ok=True)
        shutil.copy(
            "bad_castep/_LiC.castep",
            REAL_PATH
            + "/data/benchmark/results/_LiC_2core_castep{}.castep".format(
                results.get("castep_version", "xxx")
            ),
        )

        self.assertTrue(all(outputs_exist), "couldn't find output files!")
        self.assertTrue(s, "couldn't read output files!")
        self.assertLess(results["_time_estimated"], 8)

    @unittest.skipIf(
        (not CASTEP_PRESENT or not MPI_PRESENT),
        "castep or mpirun executable not found in PATH",
    )
    def test_benchmark_manycore_scf(self):
        """ Test the time taken to perform a set number of SCF steps
        on many cores.

        """
        from os import makedirs

        seed = "_LiC.res"
        shutil.copy(REAL_PATH + "data/structures/LiC.res", "_LiC.res")

        cell_dict, s = cell2dict(
            REAL_PATH + "/data/benchmark/LiC_scf/LiC_scf.cell",
            verbosity=VERBOSITY,
            db=False,
        )
        self.assertTrue(s)
        param_dict, s = param2dict(
            REAL_PATH + "/data/benchmark/LiC_scf/LiC_scf.param",
            verbosity=VERBOSITY,
            db=False,
        )
        self.assertTrue(s)

        shutil.copy(REAL_PATH + "data/pspots/Li_00PBE.usp", ".")
        shutil.copy(REAL_PATH + "data/pspots/C_00PBE.usp", ".")
        with self.assertRaises(CalculationError):
            ComputeTask(
                ncores=NCORES,
                nnodes=None,
                node=None,
                res=seed,
                param_dict=param_dict,
                cell_dict=cell_dict,
                verbosity=0,
                executable=EXECUTABLE,
                start=True,
            )

        outputs_exist = [
            isfile("bad_castep/_LiC.res"),
            isfile("bad_castep/_LiC.castep"),
        ]

        results, s = castep2dict("bad_castep/_LiC.castep", db=False)
        makedirs(REAL_PATH + "/data/benchmark/results", exist_ok=True)
        shutil.copy(
            "bad_castep/_LiC.castep",
            REAL_PATH
            + "/data/benchmark/results/_LiC_{}core_castep{}.castep".format(
                results.get("num_mpi_processes", 0),
                results.get("castep_version", "xxx"),
            ),
        )

        self.assertTrue(all(outputs_exist), "couldn't find output files!")
        self.assertTrue(s, "couldn't read output files!")
        print(results["_time_estimated"])
        benchmark_data = {2: 2 * 7.4, 4: 4 * 4.0, 12: 16.8, 14: 22.4, 18: 23.4}
        warnings.warn(
            RuntimeWarning(
                "Run took {} s with {} MPI processes, with cumulative CPU time of {:.1f} s. Benchmark data\n = {}".format(
                    results["_time_estimated"],
                    results["num_mpi_processes"],
                    results["_time_estimated"] * results["num_mpi_processes"],
                    benchmark_data,
                )
            )
        )


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python
import unittest
from os.path import realpath

# grab abs path for accessing test data
REAL_PATH = "/".join(realpath(__file__).split("/")[:-1]) + "/"


class CursorUtilTest(unittest.TestCase):
    """ Tests cursor util functions. """

    def test_guess_prov(self):
        from matador.utils.cursor_utils import get_guess_doc_provenance

        sources = []
        answers = []

        sources.append("/my/stupid/long/path/KSnP-asdf123.castep")
        answers.append("AIRSS")

        sources.append(["/my/stupid/long/path/KSnP-asdf123.castep"])
        answers.append("AIRSS")

        sources.append(["/my/ICSD-CollCode678321/stupid/long/path/KSnP-asdf123.castep"])
        answers.append("AIRSS")

        sources.append(
            ["KSnP-asdf123.res", "KSnP-ICSD-CollCode123123.cell", "OQMD_11111.param"]
        )
        answers.append("AIRSS")

        sources.append(["Na-Collo.res"])
        answers.append("ICSD")

        sources.append(["OQMD_64572"])
        answers.append("OQMD")

        sources.append(["PK-BiK-OQMD_647548-CollCode55065.castep"])
        answers.append("SWAPS")

        sources.append(["KSnP-GA-abcdef-5x101.res"])
        answers.append("GA")

        sources.append(["/u/fs1/swaps+known/LiP-KSnP-GA-abcdef-5x101.res"])
        answers.append("SWAPS")

        sources.append(
            ["/u/fs1/swaps+known/LiP-CollCode10101-swap-KSnP-GA-abcdef-5x101.res"]
        )
        answers.append("SWAPS")

        sources.append(["Ag2Bi2I8-MP-35909_300eV.castep"])
        answers.append("MP")

        sources.append(
            ["AgBiI4-spinel-Config5-DOI-10.17638__datacat.liverpool.ac.uk__240.castep"]
        )
        answers.append("DOI")

        for source, answer in zip(sources, answers):
            self.assertEqual(
                get_guess_doc_provenance(source), answer, msg="failed {}".format(source)
            )

    def test_filter_cursor(self):
        from matador.utils.cursor_utils import filter_cursor

        cursor = [{"field": int} for int in range(100)]

        filtered = filter_cursor(cursor, "field", 98)
        self.assertEqual(len(filtered), 2)

        filtered = filter_cursor(cursor, "field", [-2, 100])
        self.assertEqual(len(filtered), len(cursor))

        cursor.append({"no_field": 1})
        filtered = filter_cursor(cursor, "field", [-2, 100])
        self.assertEqual(len(filtered), len(cursor) - 1)

        cursor.append({"field": [0]})
        filtered = filter_cursor(cursor, "field", [-2, 100])
        self.assertEqual(len(filtered), len(cursor) - 2)

    def test_recursive_get_set(self):
        from matador.utils.cursor_utils import recursive_get, recursive_set

        nested_dict = {
            "source": ["blah", "foo"],
            "lattice_cart": [[1, 0, 0], [0, 1, 0], [0, 0, 1]],
            "_beef": {
                "total_energy": [1, 2, 3, 4],
                "thetas": [[1, 2, 3], [4, 5, 6], [7, 8, 9]],
                "foo": {"blah": "bloop"},
            },
        }
        self.assertEqual(recursive_get(nested_dict, ["_beef", "thetas", -1]), [7, 8, 9])
        self.assertEqual(recursive_get(nested_dict, ["_beef", "total_energy", 2]), 3)
        self.assertEqual(recursive_get(nested_dict, ["_beef", "foo", "blah"]), "bloop")
        recursive_set(nested_dict, ["_beef", "thetas", -1], [1, 2, 3])
        self.assertEqual(recursive_get(nested_dict, ["_beef", "thetas", -1]), [1, 2, 3])
        recursive_set(nested_dict, ["_beef", "total_energy", 2], 3.5)
        self.assertEqual(recursive_get(nested_dict, ["_beef", "total_energy", 2]), 3.5)
        nested_dict["_beef"]["total_energy"][2] = 3
        self.assertEqual(recursive_get(nested_dict, ["_beef", "total_energy", 2]), 3)
        nested_dict["_beef"]["foo"]["blah"] = (1, 2)
        self.assertEqual(recursive_get(nested_dict, ["_beef", "foo", "blah"]), (1, 2))

        with self.assertRaises(IndexError):
            recursive_get(nested_dict, ["_beef", "thetas", 4])
        with self.assertRaises(KeyError):
            recursive_get(nested_dict, ["_beef", "thetaz", 4])
        with self.assertRaises(KeyError):
            recursive_get(nested_dict, ["_beef", "foo", "blahp"])


if __name__ == "__main__":
    unittest.main()



# coding: utf-8
# Distributed under the terms of the MIT License.

""" This file implements a simple interface to basic PBS
functionality, mostly for monitoring walltime of jobs submitted
via PBS.

"""

from matador.compute.queueing import QueueManager


class PBSQueueManager(QueueManager):
    """ Wrapper for the PBS queueing system. """
    token = 'pbs'

    def get_ntasks(self):
        return int(self.env['PBS_TASKNUM'])

    def get_max_memory(self):
        return None

    def get_array_id(self):
        if self.env.get('PBS_ARRAYID') is not None:
            return int(self.env['PBS_ARRAYID'])
        return None

    def get_walltime(self):
        """ Query available walltime with qstat on the current job.

        Parameters:
            pbs_dict (dict): pbs env parameters to query.

        Raises:
            RuntimeError: if PBS_JOBID not present in slurm env.
            subprocess.CalledProcessError: if unable to use qstat.

        Returns:
            int: maximum allowed walltime time in seconds.

        """
        import subprocess as sp
        pbs_dict = self.env
        job_id = pbs_dict.get('PBS_JOBID')
        if job_id is not None:
            output = sp.check_output('qstat -f {}'.format(job_id), shell=True).decode('utf-8').split('\n')

        output_dict = {
            line.strip().split('=')[0].strip(): ' '.join(line.strip().split('=')[1:]).strip()
            for line in output if '=' in line
        }

        walltime = output_dict.get('Resource_List.walltime')

        hrs = 0
        if '-' in walltime:
            days = int(walltime.split('-')[0])
            walltime = walltime.split('-')[1]
            hrs += days * 24

        hrs += int(walltime.split(':')[0])
        mins = int(walltime.split(':')[1])
        secs = int(walltime.split(':')[2])
        walltime_in_seconds = (60 * hrs + mins) * 60 + secs

        return walltime_in_seconds



#!/usr/bin/env python
import unittest
from os.path import realpath
import numpy as np
from matador.fingerprints.similarity import get_uniq_cursor
from matador.fingerprints import Fingerprint
from matador.scrapers.castep_scrapers import res2dict

REAL_PATH = "/".join(realpath(__file__).split("/")[:-1]) + "/"


class SimilarityFilterTest(unittest.TestCase):
    """ Test similarity filter. """

    def test_icsd_priority(self):
        test_docs = []
        i = 0
        while i < 10:
            test_doc, _ = res2dict(REAL_PATH + "data/KP_primitive.res", db=False)
            test_doc["text_id"] = ["primitive", "cell"]
            test_docs.append(test_doc)
            i += 1

        uniq_inds, _, _, _ = get_uniq_cursor(test_docs)
        self.assertEqual(uniq_inds, [0])

        test_docs[6]["source"] = ["KP-CollCode999999.res"]
        test_docs[6]["icsd"] = 999999
        test_docs[6]["text_id"] = ["keep", "this"]

        uniq_inds, _, _, _ = get_uniq_cursor(
            test_docs, **{"dr": 0.1, "gaussian_width": 0.1}
        )
        self.assertEqual(uniq_inds, [6])

    def test_k3p_uniq_default(self):
        cursor, _ = res2dict(REAL_PATH + "data/K3P_uniq/*.res")
        cursor = sorted(cursor, key=lambda x: x["enthalpy_per_atom"])
        uniq_inds, _, _, _ = get_uniq_cursor(cursor)
        filtered_cursor = [cursor[ind] for ind in uniq_inds]
        self.assertEqual(len(cursor), 10)
        self.assertEqual(len(filtered_cursor), 5)
        found = []
        correct_structures = [
            "K3P-OQMD_4786-CollCode25550",
            "K3P-mode-follow-swap-Na3N-OQMD_21100-CollCode165992",
            "KP-fvsqdf",
            "PK-NNa3-OQMD_21100-CollCode165992",
            "KP-yzcni8",
        ]
        for struct in correct_structures:
            for doc in filtered_cursor:
                if struct in doc["source"][0]:
                    found.append(True)
                    break
            else:
                found.append(False)

        if not all(found):
            print([doc["source"][0] for doc in filtered_cursor])

        self.assertTrue(all(found))

    def test_volume_rescale(self):
        import numpy as np

        test_doc, success = res2dict(REAL_PATH + "data/KP_primitive.res", db=False)
        self.assertTrue(success)
        test_docs = []
        rescale = np.linspace(0.1, 10, 8)
        lattice = np.asarray(test_doc["lattice_abc"])
        for val in rescale:
            test_docs.append(test_doc)
            test_docs[-1]["lattice_abc"] = lattice
            test_docs[-1]["lattice_abc"][0] *= val
            test_docs[-1]["lattice_abc"] = test_docs[-1]["lattice_abc"].tolist()
        uniq_inds, _, _, _ = get_uniq_cursor(test_docs)
        self.assertEqual(uniq_inds, [0])

    def test_uniq_filter_with_hierarchy(self):
        import glob

        files = glob.glob(REAL_PATH + "data/uniqueness_hierarchy/*.res")
        cursor = [res2dict(f)[0] for f in files]
        cursor = sorted(cursor, key=lambda x: x["enthalpy_per_atom"])[0:10]
        uniq_inds, _, _, _ = get_uniq_cursor(
            cursor,
            sim_tol=0.08,
            energy_tol=0.05,
            projected=True,
            **{"dr": 0.01, "gaussian_width": 0.1}
        )
        filtered_cursor = [cursor[ind] for ind in uniq_inds]
        self.assertEqual(len(uniq_inds), 2)
        self.assertEqual(len(filtered_cursor), 2)
        self.assertTrue(
            "KP-NaP-OQMD_2817-CollCode14009" in filtered_cursor[0]["source"][0]
        )
        self.assertTrue("KP-NaP-CollCode421420" in filtered_cursor[1]["source"][0])

    def test_uniq_filter_with_hierarchy_2(self):
        cursor, f_ = res2dict(REAL_PATH + "data/hull-LLZO/*LLZO*.res")
        cursor = sorted(cursor, key=lambda x: x["enthalpy_per_atom"])[0:10]
        uniq_inds, _, _, _ = get_uniq_cursor(
            cursor,
            sim_tol=0.1,
            energy_tol=1e10,
            projected=True,
            **{"dr": 0.01, "gaussian_width": 0.1}
        )
        filtered_cursor = [cursor[ind] for ind in uniq_inds]
        self.assertEqual(len(uniq_inds), 1)
        self.assertEqual(len(filtered_cursor), 1)
        self.assertTrue("cubic-LLZO-CollCode999999" in filtered_cursor[0]["source"][0])

    def test_double_uniqueness_hierarchy(self):
        import glob

        files = glob.glob(REAL_PATH + "data/uniqueness_hierarchy/*.res")
        files += glob.glob(REAL_PATH + "data/hull-LLZO/*LLZO*.res")
        cursor = [res2dict(f)[0] for f in files]
        cursor = sorted(cursor, key=lambda x: x["enthalpy_per_atom"])[0:10]
        uniq_inds, _, _, _ = get_uniq_cursor(
            cursor,
            sim_tol=0.1,
            energy_tol=1e20,
            projected=True,
            **{"dr": 0.01, "gaussian_width": 0.1}
        )
        filtered_cursor = [cursor[ind] for ind in uniq_inds]
        self.assertEqual(len(uniq_inds), 3)
        self.assertEqual(len(filtered_cursor), 3)
        print([doc["source"] for doc in filtered_cursor])
        self.assertTrue("cubic-LLZO-CollCode999999" in filtered_cursor[0]["source"][0])
        self.assertTrue(
            "KP-NaP-OQMD_2817-CollCode14009" in filtered_cursor[1]["source"][0]
        )
        self.assertTrue("KP-NaP-CollCode421420" in filtered_cursor[2]["source"][0])

    def test_no_overlap_retains_all_structures(self):
        import glob

        files = glob.glob(REAL_PATH + "data/uniqueness_hierarchy/*.res")
        cursor = [res2dict(f)[0] for f in files]
        uniq_inds, _, _, _ = get_uniq_cursor(
            cursor,
            sim_tol=0,
            energy_tol=1e20,
            projected=True,
            debug=True,
            **{"dr": 0.1, "gaussian_width": 0.1}
        )
        filtered_cursor = [cursor[ind] for ind in uniq_inds]
        self.assertEqual(len(filtered_cursor), len(cursor))

    def test_with_crystals(self):
        from matador.crystal import Crystal
        import glob

        files = glob.glob(REAL_PATH + "data/uniqueness_hierarchy/*.res")
        cursor = [Crystal(res2dict(f)[0]) for f in files]
        uniq_inds, _, _, _ = get_uniq_cursor(
            cursor,
            sim_tol=0,
            energy_tol=1e20,
            projected=True,
            debug=True,
            **{"dr": 0.1, "gaussian_width": 0.1}
        )
        filtered_cursor = [cursor[ind] for ind in uniq_inds]
        self.assertEqual(len(filtered_cursor), len(cursor))

    def test_with_skips(self):
        from matador.crystal import Crystal
        from matador.utils.cursor_utils import filter_unique_structures
        import glob

        files = glob.glob(REAL_PATH + "data/uniqueness_hierarchy/*.res")
        cursor = [Crystal(res2dict(f)[0]) for f in files]
        filtered_cursor = filter_unique_structures(cursor, energy_tol=0)
        self.assertEqual(len(filtered_cursor), len(cursor))

        cursor = sorted(
            [res2dict(f)[0] for f in files], key=lambda doc: doc["enthalpy_per_atom"]
        )[0:10]
        for ind, doc in enumerate(cursor):
            doc["enthalpy_per_atom"] = float(-ind)

        cursor[8]["enthalpy_per_atom"] = -5.0
        cursor[9]["enthalpy_per_atom"] = -5.0001

        filtered_cursor = filter_unique_structures(cursor, energy_tol=0.003)
        self.assertEqual(len(filtered_cursor), 8)


class TestBroadening(unittest.TestCase):
    def test_broadening_agreement(self):

        hist = np.zeros((1000))
        hist[100] = 10
        hist[800] = 20
        x = np.linspace(0, 10, 1000)
        dists = [x[100]] * 10 + [x[800]] * 20

        for btype in ["gaussian", "lorentzian"]:
            gr1 = Fingerprint._broadening_distance_dominated(hist, x, 1, btype)
            gr2 = Fingerprint._broadening_unrolled(hist, x, 1, btype)
            gr3 = Fingerprint._broadening_space_dominated(dists, x, 1, btype)

            np.testing.assert_almost_equal(gr1, gr2, decimal=10)
            np.testing.assert_almost_equal(gr2, gr3, decimal=10)



# coding: utf-8
# Distributed under the terms of the MIT License.

""" This module implements the Electrode class used for calculating
relevant electrode properties from phase diagrams.

"""

import numpy as np
from typing import List, Tuple

from matador.utils.chem_utils import get_formula_from_stoich

EPS = 1e-12


class Electrode:
    """ The Electrode class wraps phase diagrams of battery electrode
    materials, and provides useful methods for electrochemical analysis.

    Note:
        The units for relevant quantities are as follows:

            gravimetric capacity --> mAh/g
            volumetric capacity  --> mAh/cm^3
            gravimetric energy density --> Wh/kg
            volumetric energy density --> Wh/l

    Attributes:
        kind (str): either 'anode' or 'cathode'. Effects conventions
            used when computing capacities.
        ion (str): species label for the active ion.
        valence (int): number of electrons transferred per ion inserted/
            removed. Currently assumes perfect Coulombic efficiency, and
            full oxidation at each stage (e.g. Li->Li+ and Mg->Mg2+).

    """

    _kinds = ['anode', 'cathode']
    valence_data = {'Li': 1, 'Na': 1, 'K': 1,
                    'Mg': 2, 'Ca': 2}

    def __init__(self, active_ion, base_material, phases, kind='anode', valence=None):
        """ Initialise Electrode material for the given ion from a convex hull.

        Parameters:
            active_ion (str): label of species to use as active ion.
            base_formula (str): formula of the starting material (either
                "empty" anode or "full" cathode).
            phases (:obj:`QueryConvexHull`/:obj:`list` of :obj:`dict`):
                either a QueryConvexHull object for the material system
                of interest, or a list of phases.

        Keyword arguments:
            kind (str): either 'anode' or 'cathode'. This decides what
                convention to use when computing capacities.
            valence (int): the valence of the active ion, if not included
                this will be looked up for common ions.

        """

        if kind not in self._kinds:
            raise TypeError('Electrode kind must be one of {}'.format(self._kinds))
        self.kind = kind
        self.ion = active_ion

        if valence is not None:
            self.valence = valence
        else:
            self.valence = self._valence_data.get('active_ion')
        if self.valence is None:
            raise RuntimeError('Unable to lookup valence of {}, please pass it manually'
                               .format(self.ion))

        # TODO: detect different materials inside convex hull
        # or directly compute if just a list of phases

    @property
    def gravimetric_capacities(self):
        pass

    @property
    def volumetric_capacities(self):
        pass

    @property
    def voltages(self):
        pass

    @property
    def average_voltage(self):
        pass

    @property
    def gravimetric_energy_density(self):
        pass

    @property
    def volumetric_energy_density(self):
        pass

    @property
    def max_gravimetric_capacity(self):
        pass

    @property
    def max_volumetric_capacity(self):
        pass

    @property
    def voltage_curve(self):
        pass

    @property
    def pdf_vs_voltage(self):
        pass

    @property
    def pdf_vs_capacity(self):
        pass

    @property
    def pxrd_vs_voltage(self):
        pass

    @property
    def pxrd_vs_capacity(self):
        pass

    @classmethod
    def calculate_average_voltage(cls, capacities, voltages):
        """ For a given set of capacities and voltages, compute
        the average voltage during charge/discharge.

        """
        trim = None
        if np.isnan(capacities[-1]):
            trim = -1
        return np.sum(voltages[1:trim] * np.diff(capacities[:trim])) / np.max(capacities[:trim])

    @classmethod
    def _find_starting_materials(self, points, stoichs):
        """ Iterate over an array of compositions and energies to find
        the starting points of the electrode, i.e. those with zero concentration
        of the active ion.

        """

        endpoints = []
        endstoichs = []
        for ind, point in enumerate(points):
            if point[0] == 0 and point[1] != 0 and point[1] != 1:
                if not any([point.tolist() == test_point.tolist() for test_point in endpoints]):
                    endpoints.append(point)
                    endstoichs.append(stoichs[ind])

        return endpoints, endstoichs


class VoltageProfile:
    """ Container class for data associated with a voltage profile.

    Attributes:
        starting_stoichiometry: the initial stoichiometry of the electrode.
        capacities: list of gravimetric capacities in mAh/g.
        voltage: list of voltages at each capacity step.
        average_voltage: the average voltage across the full cycle.
        reactions: a list of (coefficient, formula) tuples showing the
            progression of the balanced reaction, e.g.
            `[((1, "PSn")), ((1, "LiP"), (1, "LiSn"))]`
        active_ion: species label of the active ion.

    """
    def __init__(
        self,
        starting_stoichiometry: Tuple[Tuple[str, float], ...],
        capacities: List[float],
        voltages: List[float],
        average_voltage: float,
        active_ion: str,
        reactions: List[Tuple[Tuple[float, str], ...]],
    ):
        """ Initialise the voltage profile with the given data. """

        n_steps = len(voltages)
        if any(_len != n_steps for _len in [len(capacities), len(reactions)+1]):
            raise RuntimeError(
                "Invalid size of initial arrays, capacities and voltages must have same length."
                "reactions array must have length 1 smaller than voltages"
            )

        self.starting_stoichiometry = starting_stoichiometry
        self.starting_formula = get_formula_from_stoich(starting_stoichiometry)
        self.capacities = capacities
        self.average_voltage = average_voltage
        self.voltages = voltages
        self.reactions = reactions
        self.active_ion = active_ion

    def voltage_summary(self, csv=False):
        """ Prints a voltage data summary.

        Keyword arguments:
            csv (bool/str): whether or not to write a CSV file containing the data.
                If this contains a string use this as the filename.

        """

        data_str = '# {} into {}\n'.format(self.active_ion, self.starting_formula)
        data_str += "# Average voltage: {:4.2f} V\n".format(self.average_voltage)
        data_str += '# {:^10} \t{:^10}\n'.format('Q (mAh/g)', 'Voltage (V)')

        for idx, _ in enumerate(self.voltages):
            data_str += '{:>10.2f} \t{:>10.8f}'.format(self.capacities[idx], self.voltages[idx])
            if idx != len(self.voltages) - 1:
                data_str += '\n'

        if csv:
            if isinstance(csv, str):
                fname = csv
            else:
                fname = '{}_voltages.csv'.format(self.starting_formula)
            with open(fname, 'w') as f:
                f.write(data_str)

        return 'Voltage data:\n\n{}'.format(data_str)

    def __repr__(self):
        self.voltage_summary(csv=False)



# coding: utf-8
# Distributed under the terms of the MIT License.

""" The export module provides interfaces to write matador-created
dictionaries/Crystal objects to several file formats.

Currently supported formats:

    - CASTEP's .cell and .param
    - the Protein Data Bank's .pdb
    - input files for Quantum Espresso (written as .in)
    - XCrysden's .xsf
    - the custom .res file format based on SHELX used first by AIRSS
        (https://www.mtg.msm.cam.ac.uk/Codes/AIRSS)

"""


__all__ = [
    'doc2param', 'doc2cell', 'doc2pdb', 'doc2pwscf', 'doc2res',
    'doc2xsf', 'query2files', 'doc2arbitrary'
]
__author__ = 'Matthew Evans'
__maintainer__ = 'Matthew Evans'


from matador.export.export import doc2param, doc2cell, doc2pdb
from matador.export.export import doc2pwscf, doc2res, doc2xsf, doc2arbitrary
from matador.export.export import query2files



# coding: utf-8
""" This file implements a thin wrapper to the Voronoi code
written by Can Kocer, for use with matador queries and docs.

https://bitbucket.org/can_kocer/ajm_group_voronoi_code

Requires Can's code to be on your PYTHONPATH.
"""

from matador.export import doc2res
from os import remove
from os.path import isfile


def get_voronoi_substructure(doc):
    """ Run Can's Voronoi analysis on a matador doc. """
    try:
        from Vornetclass import VoronoiNetwork
        doc2res(doc, 'Vropple.res', hash_dupe=False, overwrite=True, info=False, sort_atoms=False)
        vornet = VoronoiNetwork(filename='Vropple.res')
        vornet.computeSubStrucs()
        doc['voronoi_substruc'] = [vc.getSubStruc(use_area=False) for vc in vornet.VoronoiCells]
        if isfile('Vropple.res'):
            remove('Vropple.res')
        return doc['voronoi_substruc']
    except:
        if isfile('Vropple.res'):
            remove('Vropple.res')
        return False


def get_voronoi_points(doc, debug=False):
    """ Run Can's Voronoi analysis on a matador doc
    and return nodes, face midpoints and edge midpoints.
    """
    try:
        from Vornetclass import VoronoiNetwork
        doc2res(doc, 'Vropple.res', hash_dupe=False, overwrite=True, info=False, sort_atoms=False)
        vornet = VoronoiNetwork(filename='Vropple.res')
        if debug:
            print(vornet.struc)
        vornet.computeVorNet()
        doc['voronoi_nodes'] = vornet.getNodeFracPos()
        doc['voronoi_face_midpoints'] = vornet.getFracFaceMidpoints()
        doc['voronoi_edge_midpoints'] = vornet.getFracEdgeMidpoints()
        if isfile('Vropple.res'):
            remove('Vropple.res')
        return doc['voronoi_nodes']
    except:
        if isfile('Vropple.res'):
            remove('Vropple.res')
        return False


if __name__ == '__main__':
    from matador.query import DBQuery
    from matador.hull import QueryConvexHull
    # test with LiAs
    query = DBQuery(composition=['LiAs'], subcmd='hull')
    hull = QueryConvexHull(query, no_plot=True, hull_cutoff=0)
    hull_cursor = hull.hull_cursor
    most_lithiated = hull_cursor[-2]
    most_lithiated_substruc = get_voronoi_substructure(most_lithiated)
    other_substruc = []
    for doc in hull_cursor:
        other_substruc.append(get_voronoi_substructure(doc))
    print(other_substruc)
    print(most_lithiated_substruc)



# coding: utf-8
# Distributed under the terms of the MIT License.

""" This file implements some useful geometric functions for the
construction and manipulation of convex hulls.

"""

import warnings

import numpy as np

EPS = 1e-12


def vertices2plane(points):
    """ Convert points (xi, yi, zi) for i=1,..,3 into the
    equation of the plane spanned by the vectors v12, v13.
    For unit vectors e(i):

    v12 x v13 = n = i*e(1) + j*e(2) + k*e(3)

    and so the equation of the plane is

    i*x + j*y + k*z + d = 0.

    Parameters:
        points (list of np.ndarray): list of 3 3D numpy arrays containing
            the points comprising the vertex.

    Returns:
        callable: a function which will return the vertical distance between
            the point and the plane:

    """
    v12 = points[1] - points[0]
    v13 = points[2] - points[0]
    normal = np.cross(v12, v13)
    d = -np.sum(np.dot(normal, points[0]))
    # check other points are on the plane, to some precision
    assert np.abs(np.dot(normal, points[2]) + d) < 0 + EPS
    assert np.abs(np.dot(normal, points[1]) + d) < 0 + EPS

    def get_height_above_plane(structure):
        """ Find the z-coordinate on the plane matching
        the (x, y) coordinates of the structure, then calculate
        the difference between this z and the z of the point given.
        """
        x = structure[0]
        y = structure[1]
        z = structure[2]
        if np.abs(normal[2]) < EPS:
            warnings.warn("Normal of plane {normal} is ill-defined. Returning 0 for height above plane.")
            return 0
        z_plane = -((x*normal[0] + y*normal[1] + d) / normal[2])
        height = z - z_plane
        return height

    return get_height_above_plane


def vertices2line(points):
    """ Perform a simple linear interpolation on
    two points.

    Parameters:
        points (list of np.ndarray): list of two 2D numpy arrays.
            of form [[x1, E1], [x2, E2]].

    Returns:
        (float, float): a tuple containing the gradient and
            intercept of the line intersecting the two points.

    """
    energy_pair = [points[0][1], points[1][1]]
    comp_pair = [points[0][0], points[1][0]]
    gradient = (energy_pair[1] - energy_pair[0]) / (comp_pair[1] - comp_pair[0])
    intercept = ((energy_pair[1] + energy_pair[0]) -
                 gradient * (comp_pair[1] + comp_pair[0])) / 2

    return gradient, intercept


def is_point_in_triangle(point, triangle, preprocessed_triangle=False):
    """ Check whether a point is inside a triangle.

    Parameters:
        point (np.ndarray): 3x1 array containing the coordinates of the
            point.
        triangle (np.ndarray): 3x3 array specifying the coordinates of
            the triangle vertices.

    Keyword arguments:
        preprocessed_triangle (bool): if True, treat the input triangle
            as already processed, i.e. the array contains the inverse of
            the barycentric coordinate array.

    Returns:
        bool: whether or not the point is found to lie inside the
            triangle. If all vertices of the triangle lie on the same
            line, return False.

    """

    if not preprocessed_triangle:
        cart_planes = barycentric2cart(triangle).T
        cart_planes[-1, :] = 1
        if np.linalg.det(cart_planes) == 0:
            return False
        cart_plane_inv = np.linalg.inv(cart_planes)
    else:
        cart_plane_inv = triangle

    barycentric_structure = barycentric2cart(point.reshape(1, 3)).T
    barycentric_structure[-1, :] = 1
    plane_barycentric_structure = cart_plane_inv @ barycentric_structure

    return (plane_barycentric_structure >= 0 - EPS).all()


def barycentric2cart(structures):
    """ Convert ternary (x, y) in A_x B_y C_{1-x-y}
    to positions projected onto 2D plane.

    Input structures array is of the form:

        [
            [l(1)_0, l(2)_0, Eform_0],
            [l(1)_n, l(2)_n, Eform_n]
        ]

    where l3 = 1 - l2 - l1 are the barycentric coordinates of the point
    in the triangle defined by the chemical potentials.

    Parameters:
        structures (list of np.ndarray): list of 3D numpy arrays containing
            input points.

    Returns:
        list of np.ndarray: list of numpy arrays containing converted
            coordinates.

    """
    structures = np.asarray(structures)
    cos30 = np.cos(np.pi/6)
    cos60 = np.cos(np.pi/3)
    coords = np.zeros_like(structures)
    coords[:, 0] = structures[:, 0] + structures[:, 1] * cos60
    coords[:, 1] = structures[:, 1] * cos30
    coords[:, 2] = structures[:, -1]

    return coords


class FakeHull:
    """ Implements a thin class to mimic a ConvexHull object
    that would otherwise be undefined for two points. """
    def __init__(self):
        """ Define the used hull properties. """
        self.vertices = [0, 1]
        self.simplices = []



""" Run some importer tests directly, rather than from CLI. """

import unittest
import os
import glob
import mongomock

from matador.db.importer import Spatula
from matador.query import DBQuery

REAL_PATH = "/".join(os.path.realpath(__file__).split("/")[:-1]) + "/"
CONFIG_FNAME = None
DB_NAME = "ci_test"
ROOT_DIR = os.getcwd()
DEBUG = False
MONGO_CLIENT = mongomock.MongoClient()


@mongomock.patch(servers=(("mongo_test.com:27017"),))
class TestDatabaseImport(unittest.TestCase):
    """ Tests the Spatula class. """

    def setUp(self):
        from matador.config import load_custom_settings, SETTINGS

        SETTINGS.reset()
        SETTINGS = load_custom_settings(config_fname=CONFIG_FNAME, debug=True)
        SETTINGS["mongo"]["default_collection"] = DB_NAME
        SETTINGS["mongo"]["default_collection_file_path"] = "/data/"
        SETTINGS["mongo"]["host"] = "mongo_test.com"
        SETTINGS["mongo"]["port"] = 27017
        self.settings = SETTINGS

    def tearDown(self):
        from matador.config import SETTINGS

        SETTINGS.reset()

    def test_failed_import(self):
        """ Try to import to the default collection from some
        random folder.

        """
        os.chdir(REAL_PATH + "/data/castep_files")

        args = {"no_quickstart": True}
        with self.assertRaises(RuntimeError):
            importer = Spatula(args, settings=self.settings)
        for _file in glob.glob("spatula.*"):
            os.remove(_file)

        args = {"db": DB_NAME, "no_quickstart": True}
        with self.assertRaises(RuntimeError):
            importer = Spatula(args, settings=self.settings)
        for _file in glob.glob("spatula.*"):
            os.remove(_file)

        args = {"dryrun": True, "no_quickstart": True}
        importer = Spatula(args, settings=self.settings)
        print(importer.errors)
        self.assertEqual(importer.import_count, 0)
        self.assertEqual(importer.skipped, 0)
        self.assertEqual(importer.errors, 3)
        for _file in glob.glob("spatula.*"):
            os.remove(_file)

        args = {"force": True, "no_quickstart": True}
        importer = Spatula(args, settings=self.settings)
        self.assertEqual(importer.import_count, 3)
        self.assertEqual(importer.skipped, 0)
        self.assertEqual(importer.errors, 5)
        for _file in glob.glob("spatula.*"):
            os.remove(_file)

        args = {"force": True, "recent_only": True, "no_quickstart": True}
        importer = Spatula(args, settings=self.settings)
        # this can be fiddly and a bit system dependent, just check errors+skips are constant
        self.assertEqual(importer.skipped + importer.errors, 8)
        self.assertEqual(importer.import_count, 0)
        for _file in glob.glob("spatula.*"):
            os.remove(_file)

        query = DBQuery(db=DB_NAME, mongo_settings=self.settings)
        self.assertEqual(len(query.cursor), 3)

        query = DBQuery(
            db=DB_NAME, mongo_settings=self.settings, available_values="root_source"
        )
        self.assertEqual(len(query.cursor), 3)

        query = DBQuery(
            db=DB_NAME,
            mongo_settings=self.settings,
            root_src="Na3Zn4-swap-ReOs-OQMD_759599",
        )
        self.assertEqual(len(query.cursor), 1)

        query = DBQuery(db=DB_NAME, mongo_settings=self.settings, id="no chance")
        self.assertEqual(len(query.cursor), 0)



# coding: utf-8
# Distributed under the terms of the MIT License.

""" This file implements the base class for all "fingerprints", which
here refers to any object derived from purely structural features of a
crystal, e.g. pair distribution functions (PDF) or simulated powder
X-ray diffraction (PXRD) spectra.

"""

# TODO: wrap these broadening methods with heuristics to decide which to use

import abc
import multiprocessing as mp
import os
import time
import sys
import functools

import psutil
import numba
import numpy as np

from matador.utils.print_utils import print_notify
from matador.crystal import Crystal


class Fingerprint(abc.ABC):

    fingerprint = None
    default_key = None

    @abc.abstractmethod
    def __init__(self, doc, lazy=True, *args, **kwargs):
        pass

    @abc.abstractmethod
    def calculate(self):
        pass

    @staticmethod
    # @numba.njit
    def _broadening_space_dominated(distances, r_space, width, broadening_type='gaussian'):
        """ Add broadening to the PDF by convolving distances with
        the radial space and summing. More memory-efficient if len(r_space)
        is less than len(distances).

        Parameters:
            distances (numpy.ndarray): array of pair-wise distances.
            r_space (numpy.ndarray): radial grid
            width (float): amount of broadening.
            broadening_type (str): 'gaussian' or 'lorentzian'.

        Returns:
            gr (numpy.ndarray): the unnormalised PDF.

        """
        new_space = (np.reshape(r_space, (1, len(r_space))) -
                     np.reshape(distances, (1, len(distances))).T)
        if broadening_type == 'lorentzian':
            width /= 2
            return np.sum(1 / (1 + (new_space / width)**2), axis=0)

        return np.sum(np.exp(-(new_space / width)**2), axis=0)

    @staticmethod
    @numba.njit
    def _broadening_distance_dominated(hist, r_space, width, broadening_type='gaussian'):
        """ Add broadening to the PDF by convolving the distance histogram with
        the radial space and summing. Potentially more memory-efficient than the alternative
        implementation if len(distances) > len(r_space).

        Parameters:
            hist (numpy.ndarray): histogram of pairwise frequencies.
            r_space (numpy.ndarray): radial grid
            width (float): amount of gaussian broadening.
            broadening_type (str): 'gaussian' or 'lorentzian'.

        Returns:
            gr (numpy.ndarray): the unnormalised PDF.

        """
        new_space = (np.reshape(r_space, (1, len(r_space))) -
                     np.reshape(r_space, (1, len(r_space))).T)

        if broadening_type == 'lorentzian':
            width /= 2
            return np.sum(hist / (1 + (new_space / width)**2), axis=1)

        return np.sum(hist * np.exp(-(new_space / width)**2), axis=1)

    @staticmethod
    @numba.njit
    def _broadening_unrolled(hist, r_space, width, broadening_type='gaussian'):
        """ Add broadening to the PDF by convolving the distance histogram with
        the radial space and summing. Unrolled loop to save memory.


        Parameters:
            hist (numpy.ndarray): histogram of pairwise frequencies.
            r_space (numpy.ndarray): radial grid
            width (float): amount of gaussian broadening.
            broadening_type (str): 'gaussian' or 'lorentzian'.

        Returns:
            gr (numpy.ndarray): the unnormalised PDF.

        """
        gr = np.zeros_like(r_space)

        if broadening_type == 'lorentzian':
            width /= 2
            for ind, _ in enumerate(hist):
                if hist[ind] != 0:
                    gr += hist[ind] / (1 + ((r_space - r_space[ind]) / width)**2)

        else:
            for ind, _ in enumerate(hist):
                if hist[ind] != 0:
                    gr += hist[ind] * np.exp(-((r_space-r_space[ind]) / width)**2)

        return gr


class FingerprintFactory(abc.ABC):
    """ This class computes Fingerprint objects from a list of structures,
    using multiprocessing to perform calculations concurrently. The computed
    fingerprints are stored in each structure's dictionary under the
    default key defined by the Fingerprint objects.

    Note:
        The number of processes used to concurrency is set by the following
        hierarchy:
        SLURM_NTASKS -> OMP_NUM_THREADS -> `psutil.cpu_count(logical=False)`.

    Attributes:
        nprocs (int): number of concurrent processes to be used.

    """
    fingerprint = None
    default_key = None

    def __init__(self, cursor, required_inds=None, debug=False, **fprint_args):
        """ Compute PDFs over n processes, where n is set by either
        SLURM_NTASKS, OMP_NUM_THREADS or physical core count.

        Parameters:
            cursor (list of dict): list of matador structures
            fingerprint (Fingerprint): class to compute for each structure

        Keyword arguments:
            pdf_args (dict): arguments to pass to the fingerprint calculator
            required_inds (list(int)): indices in cursor to skip.

        """
        if required_inds is None:
            required_inds = list(range(len(cursor)))
        elif len(required_inds) == 0:
            return
        else:
            print("Skipping {} structures out of {} as no comparisons are required"
                  .format(len(cursor) - len(required_inds), len(cursor)))

        if self.fingerprint is None or self.default_key is None:
            raise NotImplementedError('Do not create FingerprintFactory directly, '
                                      'use the appropriate sub-class!')

        # create list of empty (lazy) PDF objects
        if 'lazy' in fprint_args:
            del fprint_args['lazy']

        for ind, doc in enumerate(cursor):
            if isinstance(doc, Crystal):
                doc._data.pop(self.default_key, None)
            if ind in required_inds:
                doc[self.default_key] = self.fingerprint(doc, lazy=True, **fprint_args)
            else:
                doc[self.default_key] = None

        compute_list = [doc for ind, doc in enumerate(cursor) if ind in required_inds]

        # how many processes to use? either SLURM_NTASKS, OMP_NUM_THREADS or total num CPUs
        if os.environ.get('SLURM_NTASKS') is not None:
            self.nprocs = int(os.environ.get('SLURM_NTASKS'))
            env = '$SLURM_NTASKS'
        elif os.environ.get('OMP_NUM_THREADS') is not None:
            self.nprocs = int(os.environ.get('OMP_NUM_THREADS'))
            env = '$OMP_NUM_THREADS'
        else:
            self.nprocs = psutil.cpu_count(logical=False)
            env = 'core count'
        print_notify('Running {} jobs on at most {} processes, set by {}.'
                     .format(len(required_inds), self.nprocs, env))
        self.nprocs = min(len(compute_list), self.nprocs)

        start = time.time()
        if self.nprocs == 1:
            import tqdm
            for ind, doc in tqdm.tqdm(enumerate(cursor)):
                if cursor[ind][self.default_key] is not None:
                    cursor[ind][self.default_key].calculate()
        else:
            pool = mp.Pool(processes=self.nprocs)
            fprint_cursor = []
            # for large cursors, set chunk to at most 16
            # for smaller cursors, tend to use chunksize 1 for improved load balancing
            chunksize = min(max(1, int(0.25 * len(compute_list) / self.nprocs)), 16)
            results = pool.map_async(functools.partial(_calc_fprint_pool_wrapper,
                                                       key=self.default_key),
                                     compute_list,
                                     callback=fprint_cursor.extend,
                                     error_callback=print,
                                     chunksize=chunksize)
            pool.close()
            width = len(str(len(required_inds)))
            total = len(required_inds)
            while not results.ready():
                sys.stdout.write('{done:{width}d} / {total:{width}d}  {percentage:3d}%\r'
                                 .format(width=width, done=total-results._number_left,
                                         total=total, percentage=int(100*(total-results._number_left)/total)))
                sys.stdout.flush()
                time.sleep(1)

            if len(fprint_cursor) != len(required_inds):
                raise RuntimeError('There was an error calculating the desired Fingerprint')

            fprint_ind = 0
            for ind, doc in enumerate(cursor):
                if ind in required_inds:
                    if isinstance(cursor[ind], Crystal):
                        cursor[ind]._data.pop(self.default_key, None)
                    cursor[ind][self.default_key] = fprint_cursor[fprint_ind][self.default_key]
                    fprint_ind += 1

        elapsed = time.time() - start
        if debug:
            pool.close()
            print('Compute time: {:.4f} s'.format(elapsed))
            print('Work complete!')


def _calc_fprint_pool_wrapper(doc, key=None):
    """ Evaluate Fingerprint of a structure where a lazy init of the
    doc's Fingerprint object has already been made.

    Parameters:
        doc (dict): matador structures with empty PDF

    """
    doc[key].calculate()
    return {key: doc[key]}



# coding: utf-8
# Distributed under the terms of the MIT License.

""" This submodule implements some useful auxiliary routines for use in
the other plotting functions, and some generic routines for plotting
simple (x, y) line data (e.g. pair distribution functions), or lists of
(x, y) data against some third parameter (e.g. powder x-ray spectrum vs
voltage).

"""

SAVE_EXTS = ['pdf', 'png', 'svg']
MATADOR_STYLE = '/'.join(__file__.split('/')[:-1]) + '/../config/matador.mplstyle'


def set_style(style=None):
    """ Set the matplotlib style for all future plots, manually. This
    will conflict with the context manager used by the `plotting_function`
    wrapper.

    """
    import matplotlib.pyplot as plt
    if style is None or style == 'matador':
        style = MATADOR_STYLE
    if not isinstance(style, list):
        style = [style]
    # apply multiple compound styles, if present
    for styles in style:
        plt.style.use(styles)


def plotting_function(function):
    """ Wrapper for plotting functions to safely fail on X-forwarding
    errors and handle the plot style context manager.
    """

    from functools import wraps
    from matador.utils.print_utils import print_warning, print_failure
    from matador.config import load_custom_settings

    @wraps(function)
    def wrapped_plot_function(*args, **kwargs):
        """ Wrap and return the plotting function. """
        saving = False
        result = None

        # if we're going to be saving a figure, switch to Agg to avoid X-forwarding
        try:
            for arg in args:
                if arg.savefig:
                    import matplotlib
                    # don't warn as backend might have been set externally by e.g. Jupyter
                    matplotlib.use('Agg', force=False)
                    saving = True
                    break
        except AttributeError:
            pass
        if not saving:
            if any(kwargs.get(ext) for ext in SAVE_EXTS):
                import matplotlib
                matplotlib.use('Agg', force=False)
                saving = True

        settings = load_custom_settings(kwargs.get('config_fname'), quiet=True, no_quickstart=True)
        try:
            style = settings.get('plotting', {}).get('default_style')
            if kwargs.get('style'):
                style = kwargs['style']
            if style is not None and not isinstance(style, list):
                style = [style]
            if style is None:
                style = ['matador']
            if 'matador' in style:
                for ind, styles in enumerate(style):
                    if styles == 'matador':
                        style[ind] = MATADOR_STYLE

            # now actually call the function
            set_style(style)
            result = function(*args, **kwargs)

        except Exception as exc:
            if 'TclError' not in type(exc).__name__:
                raise exc

            print_failure('Caught exception: {}'.format(type(exc).__name__))
            print_warning('Error message was: {}'.format(exc))
            print_warning('This is probably an X-forwarding error')
            print_failure('Skipping plot...')

        return result

    return wrapped_plot_function


def get_linear_cmap(colours, num_colours=100, list_only=False):
    """ Create a linear colormap from a list of colours.

    Parameters:
        colours (:obj:`list` of :obj:`str`): list of fractional RGB/hex
            values of colours

    Keyword arguments:
        num_colours (int): number of colours in resulting cmap
        list_only (bool): return only a list of colours

    Returns:
        :obj:`matplotlib.colors.LinearSegmentedColormap` or :obj:`list`:
            returns list of colours if `list_only` is True, otherwise
            :obj:`matplotlib.colors.LinearSegmentedColormap`.

    """
    import numpy as np
    from matplotlib.colors import LinearSegmentedColormap, to_rgb
    colours = [to_rgb(colour) for colour in colours]

    uniq_colours = []
    _colours = [tuple(colour) for colour in colours]
    for colour in _colours:
        if colour not in uniq_colours:
            uniq_colours.append(colour)
    _colours = uniq_colours
    linear_cmap = []
    repeat = int(num_colours / len(_colours))
    for ind, colour in enumerate(_colours):
        if ind == len(_colours) - 1:
            break
        diff = np.asarray(_colours[ind + 1]) - np.asarray(_colours[ind])
        diff_norm = diff / repeat
        for i in range(repeat):
            linear_cmap.append(np.asarray(colour) + i * diff_norm)
    if list_only:
        return linear_cmap

    return LinearSegmentedColormap.from_list('linear_cmap', linear_cmap, N=num_colours)


class XYvsZPlot:
    """ This class wraps plotting (x, y) lines against a third
    variable.

    """
    def __init__(self, xys, zs, y_scale=1.0, **kwargs):
        """ Construct plot from data.

        Parameters:
            xys (:obj:`list` of :obj:`list` or numpy.ndarray): list or
                array of data to be plotted. For N lines of M samples,
                this can be provided as an (N, M, 2) or (2, M, N) array,
                or corresponding list/sublist format.
            zs (:obj:`list`): third parameter to plot lines against. The
                y-values are rescaled relative to the maximum across all
                lines so that no lines overlap (this can be overridden
                using offset_factor keyword).

        Keyword arguments:
            y_scale (float): controls the scale factor between the
                arbitrary y-scale and the z-scale.

        """

        import numpy as np

        self.plot_kwargs = kwargs

        _xys = np.asarray(xys)
        shape = np.shape(_xys)
        if shape[0] != 2 and shape[-1] != 2:
            raise RuntimeError('Data of shape {} is not compatible with XYvsZPlot.'
                               .format(shape))
        if shape[0] == 2:
            _xys = _xys.T

        self._xs = _xys[:, :, 0]
        self._ys = _xys[:, :, 1]
        self._zs = np.asarray(zs).flatten()

        if len(self._zs) != np.shape(self._xs)[0]:
            raise RuntimeError('x/y and z data do not match in shape!')

        self._y_scale = y_scale

        self.plot(**self.plot_kwargs)

    @property
    def y_scale(self):
        return self._y_scale

    @y_scale.setter
    def y_scale(self, value):
        """ Reset the y_scale and replot. """
        self._y_scale = value
        self.plot(**self.plot_kwargs)

    def get_plot(self):
        return self.fig, self.ax

    @plotting_function
    def plot(self, *args, **kwargs):
        """ Actually plot the data and optionally save it. """

        import matplotlib.pyplot as plt

        fig = plt.figure()
        ax = fig.add_subplot(111)
        for i in range(len(self._xs)):
            ax.plot(self._xs[i, :], self._y_scale * self._ys[i, :] + self._zs[i])

        self.fig = fig
        self.ax = ax
        plt.show()



# coding: utf-8
# Distributed under the terms of the MIT License.

""" The hull module provides functionality for creating phase diagrams,
voltage curves and volume curves, either directly from a database, or
from files.

"""


__all__ = ['PhaseDiagram', 'EnsembleHull', 'QueryConvexHull', 'HullDiff', 'diff_hulls']
__author__ = 'Matthew Evans'
__maintainer__ = 'Matthew Evans'


from matador.hull.hull import QueryConvexHull
from matador.hull.phase_diagram import PhaseDiagram
from matador.hull.hull_ensemble import EnsembleHull
from matador.hull.hull_diff import HullDiff, diff_hulls
